"""
Whad up ?
"""
import sys
from html import escape
from typing import List

# Helpers
from prompt_toolkit import print_formatted_text, HTML


# Whad device
from whad.exceptions import WhadDeviceNotFound, WhadDeviceNotReady
from whad.device import Device

# Whad hub
from whad.hub.ble import Commands as BleCommands
from whad.hub.dot15d4 import Commands as Dot15d4Commands
from whad.hub.esb import Commands as ESBCommands
from whad.hub.unifying import Commands as UnifyingCommands
from whad.hub.phy import Commands as PhyCommands
from whad.hub.discovery import Domain, Capability


DOMAINS = {
    Domain.Phy: "Physical Layer",
    Domain.ANT: "ANT",
    Domain.ANT_FS: "ANT FS",
    Domain.ANT_Plus: "ANT+",
    Domain.BtClassic: "Bluetooth Classic",
    Domain.BtLE: "Bluetooth LE",
    Domain.Esb: "Enhanced ShockBurst",
    Domain.LogitechUnifying: "Logitech Unifying",
    Domain.Mosart: "Mosart",
    Domain.SixLowPan: "6LowPan",
    Domain.Dot15d4: "802.15.4"
}

CAPABILITIES = {
    Capability.Scan: "can scan devices",
    Capability.Hijack: "can hijack communication",
    Capability.Hook: "can hook packets",
    Capability.Inject: "can inject packets",
    Capability.Jam: "can jam communications",
    Capability.SimulateRole: "can simulate a role in a communication",
    Capability.Sniff: "can sniff data",
    Capability.NoRawData: "can not read/write raw packet"
}

BLE_COMMANDS = {
    BleCommands.SetBdAddress: "SetBdAddress: can set BD address",
    BleCommands.SniffAdv: "SniffAdv: can sniff advertising PDUs",
    BleCommands.JamAdv: "JamAdv: can jam advertising PDUs",
    BleCommands.JamAdvOnChannel: "JamAdvOnChannel: can jam advertising PDUs on a single channel",
    BleCommands.ReactiveJam: "ReactiveJam: can reactively jam PDU on a single channel",
    BleCommands.SniffConnReq: "SniffConnReq: can sniff a new connection",
    BleCommands.SniffAccessAddress: "SniffAccessAddress: can detect active connections",
    BleCommands.SniffActiveConn: "SniffActiveConn: can sniff an active connection",
    BleCommands.JamConn: "JamConn: can jam an active connection",
    BleCommands.ScanMode: "ScanMode: can scan devices",
    BleCommands.AdvMode: "AdvMode: can advertise as a BLE device",
    BleCommands.SetAdvData: "SetAdvData: can set advertising PDU details",
    BleCommands.CentralMode: "CentralMode: can act as a Central device",
    BleCommands.ConnectTo: "ConnectTo: can initiate a BLE connection",
    BleCommands.SendRawPDU: "SendRawPDU: can send a raw PDU",
    BleCommands.SendPDU: "SendPDU: can send a PDU",
    BleCommands.Disconnect: "Disconnect: can terminate an active connection (in Central mode)",
    BleCommands.PeripheralMode: "PeripheralMode: can act as a peripheral",
    BleCommands.Start: "Start: can start depending on the current mode",
    BleCommands.Stop: "Stop: can stop depending on the current mode",
    BleCommands.SetEncryption: "SetEncryption: can enable encryption during a connection",
    BleCommands.HijackMaster: "HijackMaster: can hijack the Master role in an active connection",
    BleCommands.HijackSlave: "HijackSlave: can hijack the Slave role in an active connection",
    BleCommands.HijackBoth: (
        "HijackBoth: can hijack the Master and the Slave role in an active connection"
    ),
    BleCommands.PrepareSequence: (
        "PrepareSequence: can prepare a sequence of packets and associate a trigger"
    ),
    BleCommands.TriggerSequence: (
        "TriggerSequence: can manually trigger the transmission of a sequence of packets"
    ),
    BleCommands.DeleteSequence: "DeleteSequence: can delete a prepared sequence of packets"

}

DOT15D4_COMMANDS = {
    Dot15d4Commands.SetNodeAddress: "SetNodeAddress: can set Node address",
    Dot15d4Commands.Sniff: "Sniff: can sniff 802.15.4 packets",
    Dot15d4Commands.EnergyDetection: "EnergyDetection: can perform energy detection scans",
    Dot15d4Commands.Jam: "Jam: can jam 802.15.4 packets",
    Dot15d4Commands.Send: "Send: can transmit 802.15.4 packets",
    Dot15d4Commands.EndDeviceMode: "EndDeviceMode: can act as an End Device",
    Dot15d4Commands.CoordinatorMode: "CoordinatorMode: can act as a Coordinator",
    Dot15d4Commands.RouterMode: "RouterMode: can act as a Router",
    Dot15d4Commands.Start: "Start: can start depending on the current mode",
    Dot15d4Commands.Stop: "Stop: can stop depending on the current mode",
    Dot15d4Commands.ManInTheMiddle: "ManInTheMiddle: can perform a Man-in-the-Middle attack",
}

ESB_COMMANDS = {
    ESBCommands.SetNodeAddress: "SetNodeAddress: can set Node address",
    ESBCommands.Sniff: "Sniff: can sniff Enhanced ShockBurst packets",
    ESBCommands.Jam: "Jam: can jam Enhanced ShockBurst packets",
    ESBCommands.Send: "Send: can transmit Enhanced ShockBurst packets",
    ESBCommands.PrimaryReceiverMode: "PrimaryReceiverMode: can act as a Primary Receiver (PRX)",
    ESBCommands.PrimaryTransmitterMode: "PrimaryReceiverMode: can act as a Primary Receiver (PTX)",
    ESBCommands.Start: "Start: can start depending on the current mode",
    ESBCommands.Stop: "Stop: can stop depending on the current mode"
}

UNIFYING_COMMANDS = {
    UnifyingCommands.SetNodeAddress: "SetNodeAddress: can set Node address",
    UnifyingCommands.Sniff: "Sniff: can sniff Logitech Unifying packets",
    UnifyingCommands.Jam: "Jam: can jam Logitech Unifying packets",
    UnifyingCommands.Send: "Send: can transmit Logitech Unifying packets",
    UnifyingCommands.LogitechDongleMode: (
        "PrimaryReceiverMode: can act as a Logitech Dongle (ESB PRX)"
    ),
    UnifyingCommands.LogitechKeyboardMode: (
        "PrimaryReceiverMode: can act as a Logitech Keyboard (ESB PTX)"
    ),
    UnifyingCommands.LogitechMouseMode: (
        "LogitechMouseMode: can act as a Logitech Mouse (ESB PTX)"
    ),
    UnifyingCommands.Start: "Start: can start depending on the current mode",
    UnifyingCommands.Stop: "Stop: can stop depending on the current mode",
    UnifyingCommands.SniffPairing: "SniffPairing: can sniff a pairing process"
}

PHY_COMMANDS = {
    PhyCommands.SetASKModulation: (
        "SetASKModulation: can use Amplitude Shift Keying modulation scheme"
    ),
    PhyCommands.SetFSKModulation: (
        "SetFSKModulation: can use Frequency Shift Keying modulation scheme"
    ),
    PhyCommands.SetGFSKModulation: (
        "SetGFSKModulation: can use Gaussian Frequency Shift Keying modulation scheme"
    ),
    PhyCommands.SetBPSKModulation: (
        "SetBPSKModulation: can use Binary Phase Shift Keying modulation scheme"
    ),
    PhyCommands.SetQPSKModulation: (
        "SetQPSKModulation: can use Quadrature Phase Shift Keying modulation scheme"
    ),
    PhyCommands.SetLoRaModulation: (
        "SetLoRaModulation: can use LoRa modulation scheme"
    ),
    PhyCommands.GetSupportedFrequencies: (
        "GetSupportedFrequencies: can return a list of supported frequency ranges"
    ),
    PhyCommands.SetFrequency: "SetFrequency: can configure a given frequency",
    PhyCommands.SetDataRate: "SetDataRate: can configure the datarate",
    PhyCommands.SetEndianness: "SetEndianness: can configure the endianness",
    PhyCommands.SetTXPower: "SetTXPower: can configure the transmission power level",
    PhyCommands.SetPacketSize: "SetPacketSize: can configure the packet size",
    PhyCommands.SetSyncWord: "SetSyncWord: can configure the synchronization word",
    PhyCommands.Sniff: "Sniff: can receive arbitrary packets",
    PhyCommands.Send: "Send: can transmit arbitrary packets",
    PhyCommands.SendRaw: "SendRaw: can transmit arbitrary IQ streams" ,
    PhyCommands.ScheduleSend: "ScheduleSend: can schedule a packet to be sent at a specific time",
    PhyCommands.Jam: "Jam: can jam a physical medium",
    PhyCommands.Monitor: "Monitor: can monitor a physical medium",
    PhyCommands.Start: "Start: can start depending on the current mode",
    PhyCommands.Stop: "Stop: can stop depending on the current mode",
}

COMMANDS = {
    Domain.BtLE: BLE_COMMANDS,
    Domain.Esb: ESB_COMMANDS,
    Domain.Dot15d4: DOT15D4_COMMANDS,
    Domain.LogitechUnifying: UNIFYING_COMMANDS,
    Domain.Phy: PHY_COMMANDS

}


def get_readable_capabilities(caps: int) -> List[str]:
    """Turn device capabilities into a readable string.

    :param caps: Capabilities
    :type caps: int
    :rtype: list
    :return: A list of readable capabilities.
    """
    capabilities = []
    for i in range(24):
        if caps & (1 << i):
            capabilities.append(CAPABILITIES[caps & (1 << i)])
    return capabilities

def get_commands_desc(domain: str, commands: int) -> List[str]:
    """Turn a domain supported commands integer into a list of
    readable command names.

    :param domain: Domain name
    :type domain: str
    :param commands: Supported commands for the specified domain
    :type commands: int
    :rtype: list
    :return: A list of readable commands supported for the specified domain.
    """
    supp_commands = []
    if domain in COMMANDS:
        for i in COMMANDS[domain].keys():
            if commands & (1 << i):
                supp_commands.append(COMMANDS[domain][i])
    return supp_commands

def info(message: str):
    """Show an escaped informative message.

    :param message: Message to show.
    :type message: str
    """
    print_formatted_text(HTML("<ansicyan>[i]</ansicyan> " + escape(message)))

def error(message):
    """Show an escaped error message.

    :param message: Message to show.
    :type message: str
    """
    print_formatted_text(HTML("<ansired>[e]</ansired> " + escape(message)))

def main():
    """Main whadup/wup function.
    """
    if len(sys.argv) >= 2:
        # Retrieve target interface
        interface = sys.argv[1]

        # Connect to target device and performs discovery

        try:
            dev = Device.create(interface)

            info("Connecting to device ...")
            dev.open()
            dev.discover()

            info("Device details")
            print('')
            print_formatted_text(
                HTML("<b> Device ID:</b> " + ":".join(
                    [f"{i:02x}" for i in dev.device_id.encode()])
                )
            )
            print_formatted_text(HTML("<b> Firmware info:</b> "))
            if len(dev.info.fw_author) > 0:
                print_formatted_text(
                    HTML(f"<b> - Author :</b> {escape(dev.info.fw_author)}")
                )
            if len(dev.info.fw_url) > 0:
                print_formatted_text(
                    HTML(f"<b> - URL :</b> {escape(dev.info.fw_url)}")
                )
            print_formatted_text(
                HTML(f"<b> - Version :</b> {escape(dev.info.version_str)}")
            )
            print('')
            try:
                info("Discovering domains ...")
                domains = {}
                for domain in dev.get_domains():
                    if domain in DOMAINS:
                        caps_val = dev.get_domain_capability(domain)
                        domains[domain] = get_readable_capabilities(caps_val)
                info("Domains discovered.")
                print('')

                for domain, caps in domains.items():
                    print_formatted_text(HTML(f" <b>This device supports {DOMAINS[domain]}:</b>"))
                    for cap in caps:
                        print(f" - {cap}")
                    print('')
                    print_formatted_text(HTML(" <b>List of supported commands:</b>"))
                    for cmd in get_commands_desc(domain, dev.get_domain_commands(domain)):
                        try:
                            command_name, command_desc = cmd.split(":")
                            print_formatted_text(
                                HTML(f"  - <b>{command_name}</b>: {command_desc[1:]}")
                            )
                        except ValueError:
                            pass

                    print('')
            except Exception:
                error("[e] An error occurred while requesting this device." +
                      "We were not able to retrieve the supported domains.")

            dev.close()
        except WhadDeviceNotFound:
            error("Device not found")
            sys.exit(1)
        except WhadDeviceNotReady:
            error("Cannot communicate with the device. Make sure it is a " +
                  "WHAD compatible device and reset it.")
        except PermissionError:
            error("Cannot access the requested device (permission error).")
    else:
        info("Available devices")
        for device in Device.list(): #print("Usage: %s [device]" % sys.argv[0])
            print_formatted_text(HTML(f"- <b>{device.interface}</b>"))
            print_formatted_text(HTML(f"  <b>Type</b>: {device.type}"))
            print_formatted_text(HTML(f"  <b>Index</b>: {device.index}"))
            print_formatted_text(HTML(f"  <b>Identifier</b>: {device.identifier}"))
            print()

if __name__ == "__main__":
    # whadup/wup launcher.
    main()



"""
This module provides the :class:`whad.zigbee.connector.coordinator.Coordinator`
connector that allows creating a ZigBee coordinator tied to a WHAD device.

This coordinator creates its own network and accept end devices to join it. It
manages the join process and share the required encryption keys.
"""
from typing import List
from scapy.packet import Packet

from whad.device import Device
from whad.dot15d4.stack import Dot15d4Stack
from whad.dot15d4.stack.mac import MACManager
from whad.exceptions import UnsupportedCapability

from .base import Zigbee
from ..stack.nwk import NWKManager
from ..stack.apl.application import ApplicationObject
from ..stack.apl.constants import LogicalDeviceType
from ..stack.nwk.network import ZigbeeNetwork

class Coordinator(Zigbee):
    """
    Zigbee Coordinator interface for compatible WHAD device.
    """

    def __init__(self, device: Device, applications: List[ApplicationObject] = None):
        """ZigBee coordinator initialization.

        :param device: WHAD device to use as coordinator
        :type device: :class:`whad.device.WhadDevice`
        :param applications: list of application objects to add to the coordinator
        :type applications: list
        """
        super().__init__(device)

        if not self.can_be_coordinator():
            raise UnsupportedCapability("Coordinator")

        # Stack initialization
        MACManager.add(NWKManager)
        self.__stack = Dot15d4Stack(self)

        # Channel initialization
        self.__channel = 11
        self.__channel_page = 0

        self.enable_reception()

        self.__stack.get_layer('apl').get_application_by_name("zdo").configuration.get("configNodeDescriptor").logical_type = LogicalDeviceType.COORDINATOR
        self.__stack.get_layer('apl').initialize()
        self._init_applications(applications)

    def _init_applications(self, applications: List[ApplicationObject]):
        """Initialize the coordinator applications.

        :param applications: list of application objects to attach to the coordinator
        :type applications: list
        """
        if applications is None:
            # If no application provided, attach a default ZCL application on endpoint 1
            app = ApplicationObject("zcl_app", 0x0104, 0x0100, device_version=0,
                                    input_clusters=[], output_clusters=[])
            self.__stack.get_layer('apl').attach_application(app, endpoint=1)
        else:
            for app in applications:
                endpoint = 1
                self.__stack.get_layer('apl').attach_application(app, endpoint=endpoint)
                endpoint += 1

    def start_network(self, channel: int = 23, ext_pan_id: int = 0x6055f90000f714e4,
                      network_key: bytes = None) -> ZigbeeNetwork:
        """Start ZigBee network: initialize the extended PAN ID for APS, set channel and
        define a default network key.

        :param channel: channel to use for this network (default: 23)
        :type channel: int, optional
        :param ext_pan_id: extended PAN ID (64 bits) value to use for this network
        :type ext_pan_id: int, optional
        :param network_key: network encryption key to use for this network
        :type network_key: bytes, optional
        :return: ZigBee network object on success, ``None`` on failure
        """
        self.__stack.get_layer('aps').database.set("apsUseExtendedPANID", ext_pan_id)
        self.__stack.get_layer('aps').database.set("apsUseChannel", channel)
        zdo = self.__stack.get_layer('apl').get_application_by_name('zdo')

        # Use a default network key if not specified
        if network_key is None:
            network_key = bytes.fromhex("ae197aa680491b06458ba5e3b4e040fe")
        zdo.security_manager.provision_network_key(network_key)

        # Start network manager
        if zdo.network_manager.startup():
            # Succeeded, return network object
            return zdo.network_manager.network
        else:
            # Failed
            return None

    def network_formation(self, pan_id: int = None, channel: int = None):
        """Form ZigBee network

        :param pan_id: short PAN ID (16 bits)
        :type pan_id: int, optional
        :param channel: channel number used by the network
        :type channel: int, optional
        :return: ZigBee network information
        """
        return self.__stack.get_layer('nwk').get_service("management").network_formation(
            pan_id = pan_id,
            channel = channel
        )

    @property
    def stack(self) -> Dot15d4Stack:
        """Current ZigBee stack
        """
        return self.__stack

    def enable_reception(self):
        """Enable coordinator mode on current channel.
        """
        self.set_coordinator_mode(channel=self.__channel)

    def set_channel(self, channel: int = 11):
        """Set current channel to ``channel``.

        :param channel: new channel number to use
        :type channel: int
        """
        self.__channel = channel
        self.enable_reception()

    def perform_ed_scan(self, channel: int = 11):
        """Perform an energy detection scan on channel.

        :param channel: channel on which energy detection scan has to be performed
        :type channel: int
        """
        if not self.can_perform_ed_scan():
            raise UnsupportedCapability("EnergyDetection")
        self.__channel = channel
        super().perform_ed_scan(channel)

    def set_channel_page(self, page: int = 0):
        """Set channel page.

        :param page: channel page to use
        :type page: int
        """
        if page != 0:
            raise UnsupportedCapability("ChannelPageSelection")
        else:
            self.__channel_page = page

    def get_channel(self) -> int:
        """Retrieve current channel number.

        :return: current channel number
        :rtype: int
        """
        return self.__channel

    def get_channel_page(self) -> int:
        """Retrieve the current channel page value

        :return: current channel page value
        :rtype: int
        """
        return self.__channel_page


    def send(self, pdu: Packet, channel: int = 11) -> bool:
        """Send pdu to network.

        :param pdu: pdu to send
        :type pdu: :class:`scapy.packet.Packet`
        :return: ``True`` on success, ``False`` otherwise
        :rtype: bool
        """
        return super().send(pdu, channel=self.__channel)

    def on_pdu(self, packet):
        """PDU handler.
        """
        if (
            hasattr(packet,"metadata") and
            hasattr(packet.metadata, "is_fcs_valid") and
            not packet.metadata.is_fcs_valid
        ):
            return

        self.__stack.on_pdu(packet)

    def on_ed_sample(self, timestamp: int, sample: int):
        """Process energy detection sample.

        :param timestamp: timestamp at which the energy detection sample has been measured
        :type timestamp: int
        :param sample: energy detection sample
        :type sample: int
        """
        self.__stack.on_ed_sample(timestamp, sample)



"""BLE Scan command handler
"""

from prompt_toolkit import print_formatted_text, HTML
from whad.cli.app import command
from whad.ble import Scanner

@command('scan')
def scan_handler(app, _):
    """scan for BLE devices

    <ansicyan><b>scan</b></ansicyan>

    This command will scan for BLE devices and show them in a list.
    """
    # We need to have an interface specified
    if app.is_piped_interface():
        app.error("This command cannot be used chained with another whad tool.")
    elif app.interface is not None:
        # Switch to BLE scan mode
        with Scanner(app.interface) as scanner:
            print_formatted_text(HTML(
                "<ansigreen> RSSI Lvl  Type  BD Address        Extra info</ansigreen>"
            ))
            try:
                for device in scanner.discover_devices():
                    # Show device
                    print(device)
            except KeyboardInterrupt:
                print('\rScan terminated by user')



"""Bluetooth Low Energy emulated device.
"""
import logging
from queue import Queue
from typing import Optional

from scapy.packet import Packet
from scapy.layers.bluetooth import L2CAP_Hdr

from whad.hub.ble import BDAddress, AddressType, AdvType

from .stack.l2cap import LlcapServer

logger = logging.getLogger(__name__)

class EmulatedDevice:
    """Properties holder for a BLE device emulated by the BleScanMock.
    """

    STATE_ADVERTISING = 0
    STATE_CONNECTED = 1

    def __init__(self, address: BDAddress, adv_data: bytes = b'', scan_data: Optional[bytes] = None):
        """Create an emulated device and its associated state.

        :param address: Device BD address
        :type address: BDAddress
        :param addr_type: Address type
        :type addr_type: AddressType
        :param adv_data: Advertising data
        :type adv_data: bytes
        :param scan_data: Scan response data
        :type scan_data: bytes
        """
        # Save device address info
        self.__address = address

        # Set advertising data and state
        if len(adv_data) > 31:
            raise ValueError()
        self.__adv_data = adv_data
        if scan_data is not None and len(scan_data) > 31:
            raise ValueError()
        self.__scan_data = scan_data
        self.__next_adv_type = "adv"

        # Set connection state
        self.__handle = None

        # Set default state as advertising
        self.__state = EmulatedDevice.STATE_ADVERTISING

        # L2CAP layer
        self.__l2cap = None

    ##
    # Device address getters
    ##

    @property
    def address(self) -> BDAddress:
        """Device BD address"""
        return self.__address

    @property
    def addr_type(self) -> AddressType:
        """Device address type"""
        return AddressType.PUBLIC if self.__address.is_public() else AddressType.RANDOM

    ##
    # Mode switching
    ##

    def set_connected(self, conn_handle: int):
        """Switch to connected mode.
        """
        self.__handle = conn_handle
        self.__state = EmulatedDevice.STATE_CONNECTED

        # Create an instance of L2CAP
        self.__l2cap = LlcapServer(self.__handle)

    def set_disconnected(self):
        """Switch back to advertising mode.
        """
        self.__handle = None
        self.__l2cap = None
        self.__state = EmulatedDevice.STATE_ADVERTISING

    def set_advertising(self):
        """Switch to advertising mode.
        """
        self.set_disconnected()

    ##
    # Advertising mode
    ##

    @property
    def adv_data(self) -> bytes:
        """Advertising data"""
        return self.__adv_data

    @property
    def scan_data(self) -> Optional[bytes]:
        """Scan response data"""
        return self.__scan_data

    def get_adv_data(self) -> Optional[tuple[int, bytes]]:
        """Return the next advertising data type and bytes.
        """
        if self.__next_adv_type == "adv":
            self.__next_adv_type = "scan"
            return (AdvType.ADV_IND, self.__adv_data)
        elif self.__next_adv_type == "scan" and self.__scan_data is not None:
            self.__next_adv_type = "adv"
            return (AdvType.ADV_SCAN_RSP, self.__scan_data)
        # Failure
        return None

    ##
    # Connected mode
    ##

    def on_pdu(self, packet: Packet) -> list[Packet]:
        """Process incoming PDU.
        """
        if self.__state == EmulatedDevice.STATE_CONNECTED:
            # Forward L2CAP data to device stack
            if self.__l2cap is not None:
                return self.__l2cap.on_pdu(packet[L2CAP_Hdr], fragment=packet.LLID == 0x01)

        # Nothing to do
        return []




# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: whad/protocol/generic.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1bwhad/protocol/generic.proto\x12\x07generic\"0\n\tCmdResult\x12#\n\x06result\x18\x01 \x01(\x0e\x32\x13.generic.ResultCode\"\x19\n\x08Progress\x12\r\n\x05value\x18\x01 \x01(\r\"\'\n\x08\x44\x65\x62ugMsg\x12\r\n\x05level\x18\x01 \x01(\r\x12\x0c\n\x04\x64\x61ta\x18\x02 \x01(\x0c\"\x1a\n\nVerboseMsg\x12\x0c\n\x04\x64\x61ta\x18\x01 \x01(\x0c\"\xd4\x01\n\x07Message\x12%\n\x06result\x18\x01 \x01(\x0e\x32\x13.generic.ResultCodeH\x00\x12(\n\ncmd_result\x18\x02 \x01(\x0b\x32\x12.generic.CmdResultH\x00\x12%\n\x08progress\x18\x03 \x01(\x0b\x32\x11.generic.ProgressH\x00\x12\"\n\x05\x64\x65\x62ug\x18\x04 \x01(\x0b\x32\x11.generic.DebugMsgH\x00\x12&\n\x07verbose\x18\x05 \x01(\x0b\x32\x13.generic.VerboseMsgH\x00\x42\x05\n\x03msg*}\n\nResultCode\x12\x0b\n\x07SUCCESS\x10\x00\x12\t\n\x05\x45RROR\x10\x01\x12\x13\n\x0fPARAMETER_ERROR\x10\x02\x12\x10\n\x0c\x44ISCONNECTED\x10\x03\x12\x0e\n\nWRONG_MODE\x10\x04\x12\x16\n\x12UNSUPPORTED_DOMAIN\x10\x05\x12\x08\n\x04\x42USY\x10\x06\x62\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'whad.protocol.generic_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _RESULTCODE._serialized_start=401
  _RESULTCODE._serialized_end=526
  _CMDRESULT._serialized_start=40
  _CMDRESULT._serialized_end=88
  _PROGRESS._serialized_start=90
  _PROGRESS._serialized_end=115
  _DEBUGMSG._serialized_start=117
  _DEBUGMSG._serialized_end=156
  _VERBOSEMSG._serialized_start=158
  _VERBOSEMSG._serialized_end=184
  _MESSAGE._serialized_start=187
  _MESSAGE._serialized_end=399
# @@protoc_insertion_point(module_scope)



"""Bluetooth Low Energy cryptographic helpers (based on Bluetooth Specification)
"""
from struct import pack

from Cryptodome.Cipher import AES
from Cryptodome.Hash import CMAC
from Cryptodome.Random import get_random_bytes
from cryptography.hazmat.primitives.asymmetric.ec import SECP256R1, \
    generate_private_key, derive_private_key, EllipticCurvePublicNumbers, \
    ECDH
from scapy.layers.bluetooth4LE import LL_ENC_REQ, LL_ENC_RSP, LL_START_ENC_REQ, \
    BTLE_CTRL, BTLE_DATA, BTLE_CONNECT_REQ
from scapy.layers.bluetooth import SM_Hdr, SM_Pairing_Request, SM_Random, \
    SM_Pairing_Response, SM_Confirm, SM_Encryption_Information, SM_Master_Identification, \
    SM_Identity_Information, SM_Identity_Address_Information, SM_Signing_Information

from whad.common.analyzer import TrafficAnalyzer
from whad.hub.ble.bdaddr import BDAddress
from whad.hub.ble import Direction as BleDirection
from whad.ble.exceptions import MissingCryptographicMaterial


def generate_random_value(bits):
    """Generate a random value of provided bit size.

    :param int bits: Number of bits (8-bit aligned) to generate.
    :return bytes: Random bytes
    """
    return get_random_bytes(int(bits/8))

def e(key, plaintext):
    """
    Implements the security function e, defined in Bluetooth Core Specification,
    [Vol 3] Part H, Section 2.2.1.
    """
    aes = AES.new(key,AES.MODE_ECB)
    return aes.encrypt(plaintext)

def em1(key, ciphertext):
    """
    Implements e{-1}, the inverse of the security function e, defined in Bluetooth
    Core Specification, [Vol 3] Part H, Section 2.2.1.
    """
    aes = AES.new(key,AES.MODE_ECB)
    return aes.decrypt(ciphertext)

def s1(key, r1, r2):
    """
    Implements the key generation function s1, defined in Bluetooth Core Specification,
    [Vol 3] Part H, Section 2.2.4.
    """
    r = r1[8:16] + r2[8:16]
    return e(key,r)

def aes_cmac(key, message):
    """
    Implements the AES-CMAC authentication function AES_CMAC, defined in Bluetooth Core
    Specification, [Vol 3] Part H, Section 2.2.5.
    """
    cmac = CMAC.new(key, ciphermod=AES)
    cmac.update(message)
    return cmac.digest()

def xor(a,b):
    """
    Implements a XOR operation applied to two bytes string.
    """
    return bytes([ai ^ bi for ai,bi in zip(a,b)])


def c1(key, r, pres, preq, iat, ia, rat, ra):
    """
    Implements the Confirm value generation function c1 defined in Bluetooth Core
    Specification, [Vol 3] Part H, Section 2.2.3.
    """
    if isinstance(ia, str):
        ia = bytes.fromhex(ia.replace(":",""))

    if isinstance(ra, str):
        ra = bytes.fromhex(ra.replace(":",""))

    p1 = pres + preq + rat + iat
    p2 = b"\x00\x00\x00\x00" + ia + ra
    res1 = e(key,xor(p1,r))
    res2 = e(key,xor(res1,p2))
    return res2

def c1m1(key, confirm, pres, preq, iat, ia, rat, ra):
    """
    Implements the inverse of the Confirm value generation function c1{-1}
    defined in Bluetooth Core Specification, [Vol 3] Part H, Section 2.2.3.
    """
    if isinstance(ia, str):
        ia = bytes.fromhex(ia.replace(":",""))

    if isinstance(ra, str):
        ra = bytes.fromhex(ra.replace(":",""))

    p1 = pres + preq + rat + iat
    p2 = b"\x00\x00\x00\x00" + ia + ra
    res = xor(em1(key,xor(em1(key,confirm),p2)),p1)
    return res


def ah(key, r):
    """
    Implements the random address hash function ah defined in Bluetooth Core Specification,
    [Vol 3] Part H, Section 2.2.3.
    """
    if isinstance(r, str):
        r = bytes.fromhex(r.replace(":",""))
    rp = 13*b"\x00" + r
    return e(key, rp)[-3:]

def f4(u, v, x, z):
    """
    Implements the Confirm value generation function f4 defined in Bluetooth Core Specification,
    [Vol 3] Part H, Section 2.2.6.
    """
    return aes_cmac(x,u + v + z)

def f5(w, n1, n2, a1, a2):
    """
    Implements the key generation function f5 defined in Bluetooth Core Specification,
    [Vol 3] Part H, Section 2.2.7.
    """
    salt = bytes.fromhex("6C888391AAF5A53860370BDB5A6083BE")
    t = aes_cmac(salt, w)
    return (
            aes_cmac(t, b"\x01" + b"\x62\x74\x6c\x65" + n1 + n2 + a1 + a2 + b"\x01\x00") +
            aes_cmac(t, b"\x00" + b"\x62\x74\x6c\x65" + n1 + n2 + a1 + a2 + b"\x01\x00")
    )

def f6(w, n1, n2, r, iocap, a1, a2):
    """
    Implements the check value generation function f6 defined in Bluetooth Core Specification,
    [Vol 3] Part H, Section 2.2.8.
    """
    return (
            aes_cmac(w, n1 + n2 + r + iocap + a1 + a2)
    )


def g2(u, v, x , y):
    """
    Implements the numeric comparison value generation function g2 defined in Bluetooth Core
    Specification, [Vol 3] Part H, Section 2.2.9.
    """
    return (
            aes_cmac(x,u+v+y)[-4:]
    )

def h6(w, key_id):
    """
    Implements the link key conversion function h6 defined in Bluetooth Core Specification,
    [Vol 3] Part H, Section 2.2.10.
    """
    return (
            aes_cmac(w,key_id)
    )


def h7(salt, w):
    """
    Implements the link key conversion function h7 defined in Bluetooth Core Specification,
    [Vol 3] Part H, Section 2.2.11.
    """
    return (
            aes_cmac(salt,w)
    )

def generate_p256_keypair(private_number=None):
    """
    Generate a P256 valid secret key and the associated public key.
    If a private number is provided, use it to derive the private key.
    Otherwise, generate it from scratch.
    """
    if private_number is None:
        private_key = generate_private_key(SECP256R1())
    else:
        private_key = derive_private_key(private_number, SECP256R1())

    return private_key, private_key.public_key()

def generate_public_key_from_coordinates(x, y):
    """
    Generate the associated public key from the X and Y coordinates on the curve.
    """
    return EllipticCurvePublicNumbers(x, y, SECP256R1()).public_key()

def generate_diffie_hellman_shared_secret(own_private_key, peer_public_key):
    """
    Generate the shared secret from a private key and the peer public key.
    """
    shared_key = own_private_key.exchange(ECDH(), peer_public_key)
    return shared_key

class LinkLayerCryptoManager:
    """
    This class implements the Link Layer encryption and authentication mechanisms.
    """

    def __init__(self, ltk, master_skd, master_iv, slave_skd, slave_iv):
        self.ltk = ltk
        self.master_skd = pack(">Q",master_skd)
        self.master_iv = pack("<L",master_iv)
        self.slave_skd = pack(">Q",slave_skd)
        self.slave_iv = pack("<L",slave_iv)
        self.master_cnt = 0
        self.slave_cnt = 0

        # Generate session key diversifier
        self.skd = self.slave_skd + self.master_skd

        # Generate initialization vector
        self.iv = self.master_iv + self.slave_iv

        # Generate session key
        self.session_key = e(self.ltk, self.skd)

    def update_slave_counter(self, new_value):
        """
        Update the slave's counter (counters must be updated to allow decryption
        and encryption).
        """
        self.slave_cnt = new_value

    def update_master_counter(self, new_value):
        """
        Update the master's counter (counters must be updated to allow decryption
        and encryption).
        """
        self.master_cnt = new_value


    def increment_slave_counter(self):
        """
        Increment the slave's counter (counters must be updated to allow decryption
        and encryption).
        """
        self.slave_cnt += 1

    def increment_master_counter(self):
        """
        Increment the master's counter (counters must be updated to allow decryption
        and encryption).
        """
        self.master_cnt += 1

    def generate_nonce(self, direction=BleDirection.MASTER_TO_SLAVE):
        """
        Generate a nonce according to the counter value, the direction and the IV.
        """
        counter = pack("I",self.master_cnt if direction == BleDirection.MASTER_TO_SLAVE else self.slave_cnt)
        direction = b"\x00" if direction == BleDirection.MASTER_TO_SLAVE else b"\x80"
        return counter + direction + self.iv

    def encrypt(self, payload, direction=BleDirection.MASTER_TO_SLAVE):
        """
        Encrypt a payload (according to the direction and payload).
        """
        header = bytes([payload[0] & 0xe3])
        nonce = self.generate_nonce(direction)
        plaintext = payload[2:]
        cipher = AES.new(self.session_key, AES.MODE_CCM, nonce=nonce, mac_len=4,
                         assoc_len=len(header))
        cipher.update(header)
        ciphertext = cipher.encrypt(plaintext)
        mic = cipher.digest()
        return payload[:2] + ciphertext + mic

    def decrypt(self, payload, direction=BleDirection.MASTER_TO_SLAVE, error_tolerance=2):
        """
        Decrypt and verify a payload (according to the direction and payload).
        """
        header = bytes([payload[0] & 0xe3])
        ciphertext = payload[2:-4]
        mic = payload[-4:]
        if direction == BleDirection.MASTER_TO_SLAVE:
            old_counter = self.master_cnt
        else:
            old_counter = self.slave_cnt

        for _ in range(error_tolerance):
            nonce = self.generate_nonce(direction)
            cipher = AES.new(self.session_key, AES.MODE_CCM, nonce=nonce, mac_len=4,
                             assoc_len=len(header))
            cipher.update(header)
            plaintext = cipher.decrypt(ciphertext)
            try:
                cipher.verify(mic)
                return (payload[:2] + plaintext, True)
            except ValueError:
                if direction == BleDirection.MASTER_TO_SLAVE:
                    self.master_cnt += 1
                else:
                    self.slave_cnt += 1
        if direction == BleDirection.MASTER_TO_SLAVE:
            self.master_cnt = old_counter
        else:
            self.slave_cnt = old_counter

        return (payload[:2] + plaintext, False)



class EncryptedSessionInitialization(TrafficAnalyzer):
    """Custom traffic analyzer to detect encrypted session initialization
    from BLE PDUs.
    """

    def __init__(self, master_skd = None, master_iv = None, slave_skd = None,
                 slave_iv = None):
        super().__init__()
        self.master_skd = master_skd
        self.master_iv = master_iv
        self.slave_skd = slave_skd
        self.slave_iv = slave_iv
        self.started = False


    def reset(self):
        super().reset()
        self.master_skd = None
        self.master_iv = None
        self.slave_skd = None
        self.slave_iv = None
        self.started = False

    def process_packet(self, packet):
        if BTLE_CTRL not in packet:
            return

        if LL_ENC_REQ in packet:
            self.trigger()
            self.master_skd = packet.skdm
            self.master_iv = packet.ivm
            self.mark_packet(packet)
        elif LL_ENC_RSP in packet:
            self.trigger()
            self.slave_skd = packet.skds
            self.slave_iv = packet.ivs
            self.mark_packet(packet)
            #self.started = True
        elif LL_START_ENC_REQ in packet or packet.opcode == 0x05:
            self.trigger()
            self.mark_packet(packet)
            self.started = True

        if self.started:
            self.complete()

    @property
    def output(self):
        return {
            "master_skd" : self.master_skd,
            "master_iv" : self.master_iv,
            "slave_skd" : self.slave_skd,
            "slave_iv" : self.slave_iv,
            "started" : self.started
        }

    @property
    def crypto_material(self):
        """Retrieve recovered cryptographic material, if any.
        """
        if (
                self.master_skd is not None and
                self.master_iv is not None and
                self.slave_skd is not None and
                self.slave_iv is not None
        ):
            return (self.master_skd, self.master_iv, self.slave_skd, self.slave_iv)

        # No crypto material
        return None

    @property
    def encryption(self) -> bool:
        """Determine if encryption is enabled or not.
        """
        return self.crypto_material is not None and self.started




class LinkLayerDecryptor:
    """Custom decryptor for BLE link-layer.
    """

    def __init__(self, *keys):
        self.keys = list(keys)
        self.managers = {}
        self.master_skd = []
        self.master_iv = []
        self.slave_skd = []
        self.slave_iv = []


    def add_crypto_material(self, master_skd, master_iv, slave_skd, slave_iv):
        """Add cryptographic material to the decryptor.
        """
        self.master_skd.append(master_skd)
        self.master_iv.append(master_iv)
        self.slave_skd.append(slave_skd)
        self.slave_iv.append(slave_iv)


    def add_key(self, key):
        """Add an encryption key to the decryptor.
        """
        if isinstance(key, str):
            if len(key) == 16:
                key = key.encode('ascii')
            else:
                try:
                    key = bytes.fromhex(key.replace(":",""))
                except ValueError:
                    return False

        if not isinstance(key, bytes) or len(key) != 16:
            return False

        if key not in self.keys:
            self.keys.append(key)
            return True
        return False

    def attempt_to_decrypt(self, packet):
        """Try to decrypt a packet based on known cryptograpgic material.
        """
        if (
                len(self.master_skd) == 0 or
                len(self.master_iv) == 0 or
                len(self.slave_skd) == 0 or
                len(self.slave_iv) == 0 or
                len(self.keys) == 0
        ):
            raise MissingCryptographicMaterial()

        if BTLE_DATA not in packet:
            return (None, False)

        if packet.len == 0 and packet.LLID == 1:
            return (None, False)

        for i, skd in enumerate(self.master_skd):
            for key in self.keys:

                if key not in self.managers:
                    manager = LinkLayerCryptoManager(key, skd,
                                                     self.master_iv[i], self.slave_skd[i],
                                                     self.slave_iv[i])
                else:
                    manager = self.managers[key]

                plaintext, success = manager.decrypt(bytes(packet)[4:-3],
                                                     direction=BleDirection.MASTER_TO_SLAVE)
                if success:
                    manager.increment_master_counter()
                    self.managers[key] = manager
                else:
                    plaintext, success = manager.decrypt(bytes(packet)[4:-3],
                                                         direction=BleDirection.SLAVE_TO_MASTER)
                    if success:
                        manager.increment_slave_counter()
                        self.managers[key] = manager

                if success:
                    decrypted_packet = BTLE_DATA(plaintext)
                    decrypted_packet.len = decrypted_packet.len - 4
                    return (decrypted_packet, True)
        return (None, False)


class IdentityResolvingKeyDistribution(TrafficAnalyzer):
    """Traffic analyzer aiming at recovering an IRK
    """

    def __init__(self):
        self.irk = None
        self.address = None
        super().__init__()

    @property
    def output(self):
        """Output state
        """
        return {
            "address" : self.address,
            "irk" : self.irk
        }

    def process_packet(self, packet):
        """Update traffic analyzer state by processing a specific packet.
        """
        if SM_Identity_Information in packet:
            self.trigger()
            self.irk = packet.irk
            self.mark_packet(packet)
        elif SM_Identity_Address_Information in packet:
            self.trigger()
            self.address = BDAddress(packet.address, random=packet.atype == "random")
            self.mark_packet(packet)
            self.complete()

    def reset(self):
        """Reset analyzer state.
        """
        super().reset()
        self.irk = None
        self.address = None

class LongTermKeyDistribution(TrafficAnalyzer):
    """Traffic analyzer to recover any long-term encryption key
    """
    def __init__(self):
        self.ediv = None
        self.rand = None
        self.ltk = None
        super().__init__()

    @property
    def output(self):
        """Analyzer output state.
        """
        return {
            "ltk" : self.ltk,
            "rand" : self.rand,
            "ediv" : self.ediv
        }

    def process_packet(self, packet):
        """Update analyzer state by processing a given packet.
        """
        if SM_Encryption_Information in packet:
            self.trigger()
            self.ltk = packet.ltk
            self.mark_packet(packet)
        elif SM_Master_Identification in packet:
            self.trigger()
            self.ediv = packet.ediv
            self.rand = packet.rand
            self.mark_packet(packet)
            self.complete()

    def reset(self):
        """Reset analyzer state.
        """
        super().reset()
        self.ediv = None
        self.rand = None
        self.ltk = None

class ConnectionSignatureResolvingKeyDistribution(TrafficAnalyzer):
    """Traffic analyzer tp recover a CSRK.
    """
    def __init__(self):
        self.csrk = None
        super().__init__()

    @property
    def output(self):
        """Output satate.
        """
        return {
            "csrk" : self.csrk
        }

    def process_packet(self, packet):
        """Update analyzer state by processing a given packet.
        """
        if SM_Signing_Information in packet:
            self.trigger()
            self.csrk = packet.csrk
            self.mark_packet(packet)
            self.complete()

    def reset(self):
        super().reset()
        self.csrk = None

class LegacyPairingCracking(TrafficAnalyzer):
    """Traffic analyzer to break legacy pairing whenever it is possible.
    """
    def __init__(
        self,
        initiator = None,
        responder = None,
        pairing_req = None,
        pairing_rsp = None,
        master_confirm = None,
        master_random = None,
        slave_confirm = None,
        slave_random = None
    ):
        super().__init__()
        self.initiator = initiator
        self.responder = responder

        self.pairing_req = pairing_req
        self.pairing_rsp = pairing_rsp
        self.master_confirm = master_confirm
        self.master_random = master_random
        self.slave_confirm = slave_confirm
        self.slave_random = slave_random

        self.tk = None
        self.stk = None

    def reset(self):
        super().reset()
        self.initiator = None
        self.responder = None

        self.pairing_req = None
        self.pairing_rsp = None
        self.master_confirm = None
        self.master_random = None
        self.slave_confirm = None
        self.slave_random = None

        self.tk = None
        self.stk = None

    def process_packet(self, packet):
        """Update analyzer state by processing a given packet
        """
        if BTLE_CONNECT_REQ in packet:
            self.trigger()
            self.initiator = BDAddress(packet.InitA, random=packet.TxAdd == 1)
            self.responder = BDAddress(packet.AdvA, random=packet.RxAdd == 1)
            self.mark_packet(packet)
        elif SM_Pairing_Request in packet:
            self.trigger()
            self.pairing_req = bytes(packet[SM_Hdr].build()) # why scapy why
            self.mark_packet(packet)

        elif SM_Pairing_Response in packet:
            self.trigger()
            self.pairing_rsp = bytes(packet[SM_Hdr].build())
            self.mark_packet(packet)

        elif SM_Confirm in packet and self.master_confirm is None:
            self.trigger()
            self.master_confirm = bytes(packet.confirm)
            self.mark_packet(packet)
        elif SM_Confirm in packet and self.master_confirm is not None:
            self.trigger()
            self.slave_confirm = bytes(packet.confirm)
            self.mark_packet(packet)
        elif SM_Random in packet and self.master_random is None:
            self.trigger()
            self.master_random = bytes(packet.random)
            self.mark_packet(packet)
        elif SM_Random in packet and self.master_random is not None:
            self.trigger()
            self.slave_random = bytes(packet.random)
            self.mark_packet(packet)

        if self.ready:
            self.complete()

    def process_connected(self, initiator, responder):
        """Process a connection event.
        """
        self.initiator = initiator
        self.responder = responder

    # Harmonize this into Traffic Analyzer class, we CAN'T run a bruteforce
    # every time a packet is processed
    @property
    def ready(self):
        """Encryption material is available
        """
        return (
            self.initiator is not None and
            self.responder is not None and
            self.pairing_req is not None and
            self.pairing_rsp is not None and
            (
                (self.master_confirm is not None or
                    self.slave_confirm is not None)
                and
                (self.master_random is not None and
                self.slave_random is not None)
            )
        )

    @property
    def output(self):
        """Output state.
        """
        keys = self.keys
        if keys is None:
            return {
                "tk" : None,
                "stk" : None
            }

        tk, stk = keys
        return {
            "tk" : tk,
            "stk" : stk
        }

    @property
    def keys(self):
        """Recovered keys
        """
        if self.master_confirm is not None and self.master_random is not None:
            rand = self.master_random
            confirm = self.master_confirm
        elif  self.slave_confirm is not None and self.slave_random is not None:
            rand = self.slave_random
            confirm = self.slave_confirm
        else:
            return None

        if self.tk is not None and self.stk is not None:
            return (self.tk, self.stk)

        for i in range(0,1000000):
            tk = pack(">IIII", 0,0,0,i)

            p1 = self.pairing_rsp[::-1] + self.pairing_req[::-1] + (
                (b"\x01" if self.responder.is_random() else b"\x00") +
                (b"\x01" if self.initiator.is_random() else b"\x00")
            )

            p2 = b"\x00\x00\x00\x00" + self.initiator.value[::-1] + self.responder.value[::-1]

            a = xor(p1, rand[::-1])
            aes = AES.new(tk, AES.MODE_ECB)
            res1 = aes.encrypt(a)
            b = xor(res1, p2)
            res2 = aes.encrypt(b)

            if res2 == confirm[::-1]:

                stk = s1(tk, self.slave_random[::-1], self.master_random[::-1])
                self.tk = tk
                self.stk = stk
                return (tk, stk)
        return None



"""WHAD server tool

This utility implements a server module, allowing to create a TCP proxy
which can be used to access a device remotely.
"""
import sys
import os
import logging

from scapy.config import conf
from scapy.themes import BrightTheme

from whad.cli.app import run_app
from whad.common.pcap import extract_pcap_metadata

from whad.device import Device
from whad.exceptions import WhadDeviceNotFound, WhadDeviceError
from whad.tools.utils import list_implemented_sniffers
from whad.tools.wsniff import WhadSniffApp

logger = logging.getLogger(__name__)

class WhadPlayApp(WhadSniffApp):
    """wsniff main CLI application class.
    """

    def __init__(self):
        """Application uses an interface and has commands.
        """
        super().__init__(
            description='WHAD play tool',
            interface=False,
            pcap_argument=True
        )

        self.add_argument(
            "--flush",
            dest="flush",
            action="store_true",

        )

        # Initialize PCAP file path.
        self.pcap_file = None

    def infer_domain_from_pcap(self):
        """Infer target domain from PCAP file.

        PCAP file `reserved1` field is used by WHAD to store the domain name
        in order to use it later for replay or injection. If this information
        is not present in the provided PCAP file, user is requested to specify
        it through command-line.
        """
        self.pcap_file = None
        index_pcap_file = None
        override_domain = False
        for i, arg in enumerate(sys.argv):
            if ".pcap" in arg:
                self.pcap_file = arg
                index_pcap_file = i
            elif arg in list_implemented_sniffers():
                override_domain = True
        if index_pcap_file is not None and not override_domain:
            if os.path.exists(self.pcap_file):
                domain = extract_pcap_metadata(self.pcap_file)
                if domain in list_implemented_sniffers():
                    sys.argv.insert(index_pcap_file + 1, domain)
                else:
                    self.error("You need to provide a domain")
                    sys.exit(1)
            else:
                self.error("PCAP file not found")
                sys.exit(1)

    def build_device_path(self):
        """Create our device path based on provided parameters.
        """
        return "pcap:" + ("flush:" if self.args.flush else "") + self.args.pcap

    def pre_run(self):
        """Pre-run operations: configure scapy theme.
        """

        # If no color is not selected, configure scapy color theme
        self.infer_domain_from_pcap()
        super().pre_run()

        if not self.args.nocolor:
            conf.color_theme = BrightTheme()

        if self.args.pcap is not None:
            self.interface = Device.create(self.build_device_path())

    def run(self):
        """Wrapper for our wplay utility, since we use a PCAP file rather than
        a compatible WHAD interface.
        """
        try:
            super().run()
        except WhadDeviceNotFound:
            self.error("PCAP file not found")
        except WhadDeviceError as err:
            # Generally raised when PCAP file is not supported
            self.error(err.message)

def wplay_main():
    """Launcher for wplay.
    """
    app = WhadPlayApp()
    run_app(app)



"""
Helpers for 802.15.4 MAC operations.
"""
import struct


def is_short_address(address):
    """
    Indicates if a given address uses the short address format (16-bit) or the extended one.
    """
    try:
        _ = struct.pack("H", address)
        return True
    except struct.error:
        return False



"""
This module provides two classes to analyze Logitech Unifying communications:
- :class:`whad.unifying.utils.analyzer.UnifyingMouseMovement` that recovers mouse movements from wireless traffic;
- :class:`whad.unifying.utils.analyzer.UnifyingKeystroke` that recovers keystrokes from plaintext and decrypted wireless traffic.
"""
import locale
import logging

from scapy.packet import Packet

from whad.common.analyzer import TrafficAnalyzer, InvalidParameter
from whad.common.converters.hid.mappings import HID_MAP
from whad.unifying.stack.constants import ClickType
from whad.unifying.hid import LogitechUnifyingMouseMovementConverter
from whad.scapy.layers.unifying import Logitech_Mouse_Payload, Logitech_Unencrypted_Keystroke_Payload, Logitech_Encrypted_Keystroke_Payload
from whad.unifying.hid import LogitechUnifyingKeystrokeConverter, HIDCodeNotFound, InvalidHIDData
from whad.unifying.crypto import LogitechUnifyingKeyDerivation

logger = logging.getLogger(__name__)

def get_default_kb_locale() -> str:
    """ Retrieve the current keyboard locale. """
    # Identify current locale based on system config.
    try:
        cur_locale = locale.getlocale()
        if cur_locale is not None:
            sys_locale, _ = cur_locale
            if sys_locale is not None and '_' in sys_locale:
                hid_locale = sys_locale.split('_')[1].lower()
                if hid_locale in HID_MAP:
                    return hid_locale

            # Display warning
            logger.warning("cannot detect system locale, default to 'fr'")

    except Exception:
        logger.warning("cannot detect system locale, default to 'fr'")
        pass

    # Return French locale if locale cannot be identified
    return "fr"

class UnifyingMouseMovement(TrafficAnalyzer):
        def reset(self):
            super().reset()
            self.x = None
            self.y = None
            self.wheel_x = None
            self.wheel_y = None
            self.button = None

        @property
        def output(self):
            return {
                "x" : self.x,
                "y" : self.y,
                "wheel_x" : self.wheel_x,
                "wheel_y" : self.wheel_y,
                "button" : self.button
            }

        def process_packet(self, packet):
            if Logitech_Mouse_Payload in packet:
                self.trigger()
                self.mark_packet(packet)
                converter = LogitechUnifyingMouseMovementConverter()
                self.x, self.y = converter.get_coordinates_from_hid_data(packet.movement)
                self.wheel_x, self.wheel_y = packet.wheel_x, packet.wheel_y
                self.button = ClickType(packet.button_mask)
                self.complete()


class UnifyingKeystroke(TrafficAnalyzer):
    """
    Logitech Unifying keystroke analyzer.

    This specific traffic analyzer parses captured Logitech Unifying
    PDUs, identifies unencrypted or decrypted keystrokes and extract
    the corresponding key based on the configured locale.
    """
    PARAMETERS = {
        "locale": get_default_kb_locale(),
    }

    def __init__(self, **kwargs):
        """Initialize Unifying keystroke analyzer."""
        super().__init__(**kwargs)

        # save pending keystrokes and recovered keystrokes
        self.__pending_keys = None
        self.__keys = None

    def set_locale(self, locale: str):
        """
        Change locale to the specified `locale`.

        :param locale: New locale to use.
        :type  locale: str
        :raise: InvalidParameter
        """
        # First, make sure the locale is valid
        if locale in HID_MAP:
            self.set_param("locale", locale)
            self.__locale = locale
        else:
            raise InvalidParameter("locale", locale)

    @property
    def output(self):
        return {
            "key" :self.__keys
        }

    def process_packet(self, packet: Packet):
        """Process an incoming packet.

        :param packet: Incoming packet to process.
        :type  packet: scapy.packet.Packet
        """
        # Extract HID data from packet, if any
        hid_data = None
        if Logitech_Unencrypted_Keystroke_Payload in packet:
            self.trigger()
            self.mark_packet(packet)
            hid_data = packet.hid_data

        if Logitech_Encrypted_Keystroke_Payload in packet:
            self.trigger()
            self.mark_packet(packet)
            if hasattr(packet, "decrypted") and packet.decrypted is not None:
                hid_data = packet.decrypted.hid_data
            else:
                hid_data = packet.hid_data

        # Process found HID data
        if hid_data is not None:
            # Empty HID report ?
            if hid_data == b"\x00" * 7:
                # Got an empty HID report, send pending keystroke if any.
                if self.__pending_keys is not None:
                    self.__keys = self.__pending_keys
                    self.__pending_keys = None
                    self.complete()
            else:
                try:
                    # Non-empty report, recover pressed key based on HID scan code and locale
                    key = LogitechUnifyingKeystrokeConverter.get_key_from_hid_data(hid_data, locale=self.__locale)
                    if len(key) > 1:
                        key = " [{}] ".format(key)

                    # If pending keystroke and a detected keystroke are different, we must report
                    # the pending keystroke as we may have missed an event (key release).
                    if self.__pending_keys is not None and key != self.__pending_keys:
                        self.__keys = self.__pending_keys
                        self.__pending_keys = None
                        self.complete()

                    # Keep pending keystroke in memory.
                    self.__pending_keys = key
                except (HIDCodeNotFound, InvalidHIDData):
                    pass

    def reset(self):
        """
        Reset this analyzer.

        :raise: InvalidParameter
        """
        super().reset()

        # Reload locale from paramater and make sure it is valid
        locale = self.get_param("locale")
        if locale in HID_MAP:
            self.__locale = locale
        else:
            raise InvalidParameter("locale", locale)

        # Reset current keypress
        self.__keys = None


analyzers = {
    "pairing_cracking" : LogitechUnifyingKeyDerivation,
    "mouse" : UnifyingMouseMovement,
    "keystroke" : UnifyingKeystroke
}




"""ZigBee utility networks caching module.

This module implements a single class, `ZigbeeNetworksCache`, that keeps track
of discovered devices and their information.
"""
from whad.zigbee.profile.network import Network
from whad.dot15d4.address import Dot15d4Address

class ZigbeeNetworksCache(object):
    """ZigBee Networks cache.
    """

    def __init__(self):
        """Initialize network cache.
        """
        self.__networks = {}

    def iterate(self):
        for network in self.__networks:
            yield self.__networks[network]

    def add(self, network: Network):
        """Add a discovered network to our cache.

        @param network Network: Network to add
        """
        self.__networks[str(Dot15d4Address(network.extended_pan_id)).lower()] = {
            'info':network,
            'discovered': False
        }

    def __getitem__(self, extended_pan_id: Dot15d4Address):
        """Return an existing network from cache.

        @param extended_pan_id str: Network extended PAN ID (i.e. '00:11:22:33:44:55:66:77')
        """

        if str(extended_pan_id).lower() in self.__networks:
            return self.__networks[str(extended_pan_id).lower()]
        else:
            # If not found, look into networks PAN ID
            for net in self.__networks:
                network = self.__networks[net]
                if network['info'].pan_id is not None and network['info'].pan_id == extended_pan_id:
                    return self.__networks[net]
            raise IndexError
    '''
    def add_profile(self, address: str, profile: PeripheralDevice):
        """Add profile to a cached device.

        @param address str: Device BD address
        @param profile PeripheralDevice: Device object (obtained when connected)
        """
        address = address.lower()
        if address in self.__devices:
            self.__devices[address.lower()]['profile'] = profile
        else:
            raise IndexError

    def get_profile(self, address: str):
        """Get cached services associated with a specific device.
        """
        address = address.lower()
        if address in self.__devices:
            # Device found, check if we have a cached profile
            return self.__devices[address]['profile']
        else:
            raise IndexError

    def mark_as_discovered(self, address):
        address = address.lower()
        if address in self.__devices:
            # Device found, check if we have a cached profile
            self.__devices[address]['discovered'] = True
        else:
            raise IndexError

    def is_discovered(self, address):
        address = address.lower()
        if address in self.__devices:
            # Device found, check if we have a cached profile
            return self.__devices[address]['discovered']
        else:
            raise IndexError
    '''



"""
Bluetooth Low Energy Tiny Stack - Attributes
"""
from struct import pack
from typing import Optional

from whad.ble.profile.attribute import UUID

class Attribute:
    """Default attribute"""

    def __init__(self, handle: int, uuid: UUID, value: bytes, end_handle : Optional[int] = None,
                 read: bool = True, write: bool = False
                ):
        """Attribute initialization."""
        self.__handle = handle
        self.__end_handle = end_handle or handle
        self.__uuid = uuid
        self.__value = value
        self.__read = read
        self.__write = write

    @property
    def handle(self) -> int:
        """Attribute handle"""
        return self.__handle

    @property
    def end_handle(self) -> int:
        """Attribute end handle (grouping)."""
        return self.__end_handle

    @property
    def uuid(self) -> UUID:
        """Attribute type uuid"""
        return self.__uuid

    @property
    def value(self) -> bytes:
        """Attribute value"""
        return self.__value
    @value.setter
    def value(self, new_value: bytes):
        """Update attribute value"""
        self.__value = new_value

    def readable(self) -> bool:
        """Check if attribute has the read permission."""
        return self.__read

    def writeable(self) -> bool:
        """Check if attribute can be written."""
        return self.__write

    def pack(self) -> bytes:
        return pack("<H", self.handle) + self.value


class PrimaryService(Attribute):
    """GATT Primary Service"""

    def __init__(self, handle: int, end_handle: int, uuid: UUID):
        """Initialize service."""
        super().__init__(handle, UUID(0x2800), uuid.packed, end_handle=end_handle)


class SecondaryService(Attribute):
    """GATT Secondary Service."""

    def __init__(self, handle: int, end_handle: int, uuid: UUID):
        """Initialize secondary service."""
        super().__init__(handle, UUID(0x2801), uuid.packed, end_handle=end_handle)


class IncludeService(Attribute):
    """GATT Included Service."""

    def __init__(self, handle: int, service_handle: int, end_handle: int, uuid: UUID):
        """Initialize included service."""
        service_value = pack("<HH", service_handle, end_handle)
        if uuid.type == UUID.TYPE_16:
            service_value += uuid.packed
        super().__init__(handle, UUID(0x2802), service_value, end_handle=end_handle)


class Characteristic(Attribute):
    """GATT Characteristic."""

    PROP_BROADCAST = 0x01
    PROP_READ = 0x02
    PROP_WRITE_WITHOUT_RESP = 0x04
    PROP_WRITE = 0x08
    PROP_NOTIFY = 0x10
    PROP_INDICATE = 0x20
    PROP_AUTH_SIGN_WRITE = 0x40
    PROP_EXTENDED = 0x80

    def __init__(self, handle: int, uuid: UUID, value_handle: int, properties: int):
        """Characteristic initizalization."""
        self.__properties = properties
        self.__value_handle = value_handle
        charac_def = pack("<BH", properties, value_handle) + uuid.packed
        super().__init__(handle, UUID(0x2803), charac_def)

    @property
    def value_handle(self) -> int:
        """Value handle."""
        return self.__value_handle

    def has_property(self, prop) -> bool:
        """Check if a property is defined for this characteristic."""
        return (prop & self.__properties) > 0

class CharacteristicValue(Attribute):
    """GATT Characteristic value."""

    def __init__(self, handle: int, uuid: UUID, value: bytes, write: bool = True, read: bool = True, 
                 write_without_resp: bool = False):
        """Characteristic value initialization."""
        self.__without_resp = write_without_resp
        # Use `write=True` because characteristic value can be written.
        super().__init__(handle, uuid, value, read=read, write=write or write_without_resp)

    def writeable_without_resp(self) -> bool:
        """Determine if the caracteristic is writeable through a WriteCommand."""
        return self.__without_resp


class ClientCharacteristicConfigurationDescriptor(Attribute):
    """GATT CCCD attribute."""

    def __init__(self, handle: int):
        """Initialize a CCCD. """
        # Descriptor value can be modified.
        super().__init__(handle, UUID(0x2902), pack("<H", 0),
                         write=True)


def find_attr_by_handle(attributes: list, handle: int) -> Attribute:
    """Find attribute from list by handle."""
    for attribute in attributes:
        if attribute.handle == handle:
            return attribute
    raise IndexError()

def find_attr_by_range(attributes: list, start_handle: int, end_handle: int) -> list[Attribute]:
    attrs = []
    for attribute in attributes:
        if attribute.handle >= start_handle and attribute.handle <= end_handle:
            attrs.append(attribute)
    return attrs

def find_attr_by_type(attributes: list[Attribute], attr_type: UUID, start_handle: int = 0, end_handle: int = 0xFFFF) -> list[Attribute]:
    """Find attribute from list by attribute type."""
    attrs = []
    for attribute in attributes:
        if attribute.handle >= start_handle and attribute.handle <= end_handle:
            if attribute.uuid == attr_type:
                attrs.append(attribute)
    return attrs

def find_charac_by_desc_handle(attributes: list[Attribute], handle: int) -> Optional[Characteristic]:
    """Find the characteristic attribute that holds the given descriptor."""
    charac = None
    for attribute in attributes:
        if isinstance(attribute, Characteristic):
            charac = attribute
        if attribute.handle == handle:
            break
    # Return found characteristic
    return charac




"""Machine module
"""
import sys
import uuid

class Machine:
    """Store data about the host machine.
    """

    # Unique ID
    unique_id = None

    @staticmethod
    def gen_unique_id() -> bytes:
        """Generate a unique ID for this machine and store it in a temporary
        standardized file. This unique is store on the filesystem and can be
        accessed by other tools to synchronize the obfuscated values.

        :rtype: bytes
        :return: Unique ID for this machine
        """
        if Machine.unique_id is None:
            # Generate and save machine seed
            Machine.unique_id = bytes(uuid.uuid4().bytes.hex(), encoding="utf-8")
            with open("/tmp/whad-machine-id", "wb") as machine_id:
                machine_id.write(Machine.unique_id)

        return Machine.unique_id

    @staticmethod
    def get_unique_id():
        """Return this machine unique ID.

        :rtype: bytes
        :return: Machine unique ID
        """
        # On Linux platforms, use /etc/machine-id as unique ID.
        if sys.platform in ("linux", "linux2"):
            try:
                with open("/etc/machine-id", "rb") as mid:
                    return mid.read().strip()
            except FileNotFoundError:
                pass
            except IOError:
                pass

        # On other systems, try to write "/tmp/whad-machine-id" to emulate the
        # behaviour of /etc/machine-id
        try:
            with open("/tmp/whad-machine-id", "rb") as machine_id:
                Machine.unique_id = machine_id.read().strip()
                return Machine.unique_id
        except FileNotFoundError:
            return Machine.gen_unique_id()
        except IOError:
            Machine.unique_id = uuid.uuid4().bytes
            return Machine.unique_id



"""
Bluetooth Low Energy
"""
import re

from whad.ble.stack.gatt import GattClient, GattServer
from whad.common.triggers import ManualTrigger, ConnectionEventTrigger, ReceptionTrigger
from whad.hub.ble.bdaddr import BDAddress
from whad.hub.ble.chanmap import ChannelMap

from whad.ble.profile.attribute import UUID
from whad.ble.profile import (
    GenericProfile, Profile, Characteristic, Descriptor, ClientCharacteristicConfig,
    ReportReference, UserDescription, Service, PrimaryService, SecondaryService,
    IncludeService, read, write, written, subscribed, unsubscribed,

    CharacteristicDescriptor, CharacteristicUserDescriptionDescriptor, ReportReferenceDescriptor
)
from whad.ble.profile.advdata import (
    AdvDataFieldList, AdvFlagsField, AdvDataField, AdvCompleteLocalName,
    AdvManufacturerSpecificData, AdvShortenedLocalName, AdvTxPowerLevel,
    AdvDataFieldListOverflow, AdvURI, AdvLeRole, AdvAppearance, AdvUuid128List,
    AdvUuid16List, AdvDataError
)
from whad.ble.profile.services import BatteryService, DeviceInformationService, GapService, HeartRateService
from whad.ble.connector.base import BLE
from whad.ble.connector import (
    Central, Peripheral, Sniffer, Hijacker, Injector, Scanner, PeripheralClient, Scanner
)
from whad.ble.utils.phy import PHYS

def is_bdaddr_valid(bd_addr):
    return re.match('^([0-9a-fA-F]{2}:){5}[0-9a-fA-F]{2}$',bd_addr)

__all__ = [
    # Basic types for BLE
    'BLE',
    'ChannelMap',
    'GattClient',
    'GattServer',
    'BDAddress',
    'UUID',
    'Profile',
    'read',
    'write',
    'written',
    'subscribed',
    'unsubscribed',
    'BatteryService',
    'GapService',
    'HeartRateService',
    'DeviceInformationService',
    'Characteristic',
    'Descriptor',
    'ClientCharacteristicConfig',
    'ReportReference',
    'UserDescription',
    'Service',
    'PrimaryService',
    'SecondaryService',
    'IncludeService',

    # Advertising data classes
    'AdvDataFieldList',
    'AdvFlagsField',
    'AdvDataField',
    'AdvCompleteLocalName',
    'AdvManufacturerSpecificData',
    'AdvShortenedLocalName',
    'AdvTxPowerLevel',
    'AdvDataFieldListOverflow',
    'AdvURI',
    'AdvLeRole',
    'AdvAppearance',
    'AdvUuid128List',
    'AdvUuid16List',
    'AdvDataError',

    # Connectors
    'Central',
    'Peripheral',
    'PeripheralClient',
    'Sniffer',
    'Hijacker',
    'Injector',
    'Scanner',

    # Misc.
    'PHYS',
    'ConnectionEventTrigger',
    'ManualTrigger',
    'ReceptionTrigger',

    # Old classes to remove later
    'GenericProfile',
    'CharacteristicDescriptor',
    'CharacteristicUserDescriptionDescriptor',
    'ReportReferenceDescriptor',
]



from whad.zigbee.stack.apl.zcl import ZCLCluster, ZCLClientCluster, ZCLServerCluster

class OnOffClient(ZCLClientCluster):
    """
    Zigbee Cluster Library OnOff Client Cluster.
    """
    def __init__(self):
        super().__init__(cluster_id=0x0006)

    @ZCLCluster.command_generate(0x00, "Off")
    def off(self):
        command = b""
        self.send_command(command)
        status = self.wait_response()
        return status == 0

    @ZCLCluster.command_generate(0x01, "On")
    def on(self):
        command = b""
        self.send_command(command)
        status = self.wait_response()
        return status == 0

    @ZCLCluster.command_generate(0x02, "Toggle")
    def toggle(self):
        command = b""
        self.send_command(command)
        status = self.wait_response()
        return status == 0

class OnOffServer(ZCLServerCluster):
    """
    Zigbee Cluster Library OnOff Server Cluster.
    """
    OnOff: (0x0000, ["read", "report", "scene"]) = 0
    GlobalSceneControl: (0x4000, ["read"]) = 1
    OnTime: (0x4001, ["read", "write"]) = 0
    OffWaitTime: (0x4002, ["read", "write"]) = 0

    def __init__(self):
        super().__init__(cluster_id=0x0006)


    @ZCLCluster.command_receive(0x00, "Off")
    def on_off(self, command):
        self.OnOff = 0
        print("[i] Off")
        # default response ?

    @ZCLCluster.command_receive(0x01, "On")
    def on_on(self, command):
        self.OnOff = 1
        print("[i] On")
        # default response ?

    @ZCLCluster.command_receive(0x02, "Toggle")
    def on_toggle(self, command):
        self.OnOff = 1 - self.OnOff
        print("[i] Toggle")
        # default response ?



"""
WHAD default bridge module.

This module provides multiple classes to implement a transparent bridge between
two connectors.
"""
import logging
from threading import Event

from whad.hub.message import AbstractPacket, HubMessage

from .connector import Connector

logger = logging.getLogger(__name__)

class BridgeIfaceWrapper(Connector):
    """Interface bridging wrapper used by our default `Bridge` class
    to wrap an existing and initialized WHAD device.
    """

    def __init__(self, device, processor):
        super().__init__(device)
        self.__processor = processor

    def send_message(self, message, keep=None):
        """Forward message to the underlying interface.
        """
        logger.debug("[PacketProcIfaceWrapper][%s] prepare callback for message %s",
                     self.device.interface, message)
        message.callback(self.on_message_sent)
        logger.debug("[PacketProcIfaceWrapper] send_message: %s", message)
        super().send_message(message, keep=keep)

    def on_message_sent(self, message, status):
        """Called when a message has been sent to the target interface.
        """
        if status == 0:
            self.__processor.on_message_sent(self, message)
        else:
            # TODO: implement or remove on_message_error()
            self.__processor.on_message_error(self, message, status)


    def on_disconnection(self):
        """Notify bridge on disconnection.
        """
        self.__processor.on_disconnect(self)

    def unlock(self, dispatch_callback=None):
        """Unlock connector and dispatch pending PDUs.

        :param  dispatch_callback: PDU dispatch callback that overrides the
                                   internal dispatch routine
        :type   dispatch_callback: callable
        """
        super().unlock(self.dispatch_locked_pdus)

    def dispatch_locked_pdus(self, pdu):
        """Process locked pdus.
        """
        # convert packet to message
        msg = self.hub.convert_packet(pdu)
        #logger.debug("[PacketProcIfaceWrapper] process locked pdu %s", msg)
        self.on_any_msg(msg)

    def on_any_msg(self, message):
        self.__processor.on_any_msg(self, message)

    def on_generic_msg(self, message):
        pass

    def on_discovery_msg(self, message):
        pass

    def on_domain_msg(self, domain, message):
        pass

    def on_packet(self, packet):
        pass

    def on_event(self, event):
        pass


class Bridge:
    """WHAD default interface bridge.

    This class implements a bridge between two connectors that offers
    the possibility to filter ingress and egress packets. By default,
    every packet is forward from one interface to the other.
    """

    def __init__(self, input_connector, output_connector):
        """Initialize our packet processor.
        """
        # Stopped event
        self.__stopped = Event()

        # Save our input connector and its device
        self.__in = input_connector
        self.__in_device = self.__in.device
        self.__in_disconnected = False

        # Save our output cnnector and its device
        self.__out = output_connector
        self.__out_device =self.__out.device
        self.__out_disconnected = False

        # Disable message queue filters
        self.__in_device.set_queue_filter(None)
        self.__out_device.set_queue_filter(None)

        # Bridge our two interfaces with our own connectors.
        # This will replace each device connector with our own and avoid
        # any other packet processing.
        self.__in_wrapper = BridgeIfaceWrapper(self.__in_device, self)
        self.__out_wrapper = BridgeIfaceWrapper(self.__out_device, self)

        # Unlock connectors, if locked.
        self.unlock()

    def detach(self):
        """Detach BridgeIfaceWrappers from bridge's devices.
        """
        self.__in_device.set_connector(self.__in)
        self.__out_device.set_connector(self.__out)


    @property
    def input(self) -> Connector:
        """Get the input connector

        :return: Input connector
        :rtype: Connector
        """
        return self.__in

    @property
    def output(self) -> Connector:
        """Get the output connector

        :return: Output connector
        :rtype: Connector
        """
        return self.__out

    @property
    def input_wrapper(self) -> Connector:
        """Get the internal connector for input
        """
        return self.__in_wrapper
    
    @property
    def output_wrapper(self) -> Connector:
        """Get the internal connector for output
        """
        return self.__out_wrapper

    def on_message_sent(self, _, __):
        """Called whenever a wrapped interface has successfully sent a message
        to the hardware device.
        """

        # Is the output interface disconnected and input interface done ?
        if self.__out_disconnected and not self.__in_wrapper.busy():
            logger.debug("[bridge::msg_sent] We are done with packets !")
            # We are done processing packets, bridge stops.
            self.__stopped.set()

        # Is the input interface disconnected and output interface done ?
        elif self.__in_disconnected and not self.__out_wrapper.busy():
            logger.debug("[bridge::msg_sent] We are done with packets !")
            # We are done processing packets, bridge stops.
            self.__stopped.set()

    def on_disconnect(self, wrapper: BridgeIfaceWrapper):
        """When a wrapper disconnects, stop bridge.
        """
        if wrapper == self.__in_wrapper:
            # Input interface has disconnected, we won't receive messages from
            # this interface anymore. We still need to make sure the output
            # interface has processed and sent all the messages we sent to it
            # before closing the bridge.
            logger.debug("[bridge][%s] interface has just disconnected",
                         self.__in_device.interface)
            self.__in_disconnected = True

            # If the output wrapper is not busy, we are done processing packets.
            if not self.__out_wrapper.busy():
                logger.debug("[bridge] We are done with packets !")
                self.__stopped.set()

        elif wrapper == self.__out_wrapper:
            logger.debug("[bridge][%s] interface has just disconnected",
                self.__out_device.interface)
            self.__out_disconnected = True

            # If the input wrapper is not busy, we are done processing packets.
            if not self.__in_wrapper.busy():
                logger.debug("[bridge] We are done with packets !")
                self.__stopped.set()


    def on_any_msg(self, wrapper, message: HubMessage):
        """Callback method for any message.

        :param wrapper: Calling wrapper object
        :type wrapper: BridgeIfaceWrapper
        :param message: Incoming message
        :type message: HubMessage
        """
        logger.debug("bridge::on_any_msg - %s", message)
        if wrapper == self.__in_wrapper:
            self.on_outbound(message)
        elif wrapper == self.__out_wrapper:
            self.on_inbound(message)
        else:
            logger.error("on_any_msg() called by an unknown wrapper (%s)", wrapper)

    def on_inbound(self, message: HubMessage):
        """Inbound message hook

        This hook is called whenever a message is received on the input
        connector and about to be relayed to the output connector.

        :param message: Message to process
        :type message: HubMessage
        """
        if message is not None:
            # Monitor packet if required
            if issubclass(message, AbstractPacket):
                packet = message.to_packet()
                if packet is not None:
                    self.__in.monitor_packet_rx(packet)
                    self.__out.monitor_packet_tx(packet)

            # Forward message
            message.callback(self.on_message_sent)
            self.__in.send_message(message)

    def on_outbound(self, message: HubMessage):
        """Outbound message hook.

        This hook is called whenever a message is received on the output
        connector and about to be relayed to the input connector.

        :param message: Message to process
        :type message: HubMessage
        """
        if message is not None:
            # Monitor packet if required
            if issubclass(message, AbstractPacket):
                packet = message.to_packet()
                if packet is not None:
                    self.__out.monitor_packet_tx(packet)
                    self.__in.monitor_packet_rx(packet)

            message.callback(self.on_message_sent)
            self.__out.send_message(message)

    def unlock(self):
        """Unlock bridge's interfaces.
        """
        # Unlock connector, causing packets to be sent to the output connector
        if self.__in.is_locked():
            logger.debug("[bridge] %s is locked, unlocking it", self.__in.device.interface)
            self.__in.unlock(dispatch_callback=self.dispatch_pending_input)
        if self.__out.is_locked():
            logger.debug("[bridge] %s is locked, unlocking it", self.__out.device.interface)
            self.__out.unlock(dispatch_callback=self.dispatch_pending_output)

    def dispatch_pending_output(self, message: HubMessage):
        """Forward pending output messages to input.
        """
        packet = message.to_packet()
        if packet is not None:
            self.output.monitor_packet_rx(packet)
            self.input.monitor_packet_tx(packet)
        self.input.send_message(message)

    def dispatch_pending_input(self, message: HubMessage):
        """Forward pending input messages to output.
        """
        packet = message.to_packet()
        if packet is not None:
            self.output.monitor_packet_tx(packet)
            self.input.monitor_packet_rx(packet)
        self.output.send_message(message)

    def join(self):
        """Wait for bridge termination. Bridge termination is triggered when
        at least one of the bridge's interface is disconnected and every pending
        messages forwarded.
        """
        return self.__stopped.wait()



"""WHAD generic injection configuration for BLE.

This module defines the :class:`whad.ble.injecting.InjectionConfiguration`
used by `winject` to determine the parameters required for packet injection.
"""

from dataclasses import dataclass
from whad.ble.sniffing import ConnectionConfiguration


@dataclass
class InjectionConfiguration:
    """
    Configuration for the Bluetooth Low Energy injector.

    :param inject_to_master: inject packet to master in a synchronized connection
    :param inject_to_slave: inject packet to slave in a synchronized connection
    :param raw: inject a packet directly
    :param synchronize: synchronize before injection (s)
    :param active_connection: enable and configure existing connection synchronization
    :param channel: select the channel to sniff (c)
    :param filter: display only the packets matching the filter BD address (m)
    """
    raw : bool = False
    inject_to_slave : bool = False
    inject_to_master : bool = False
    synchronize : bool = False
    active_connection : ConnectionConfiguration = None
    channel : int = None
    filter : str = "FF:FF:FF:FF:FF:FF"



from whad.dot15d4.stack.mac.constants import MACAddressMode, MACNetworkType

class Dot15d4PANNetwork:
    """
    Represents a 802.15.4 PAN network.
    """
    def __init__(self, beacon, channel_page, channel):
        try:
            self.coord_addr_mode = MACAddressMode(beacon.fcf_srcaddrmode)
        except:
            self.coord_addr_mode = None
        self.coord_pan_id = beacon.src_panid
        self.coord_addr = beacon.src_addr
        self.channel = channel
        self.channel_page = channel_page
        self.link_quality = beacon.metadata.lqi if hasattr(beacon, "metadata") and hasattr(beacon.metadata, "lqi") else 255
        self.beacon_order = beacon.sf_beaconorder
        self.superframe_order = beacon.sf_sforder
        self.final_capslot = beacon.sf_finalcapslot
        self.batt_life_extend = beacon.sf_battlifeextend
        self.pan_coord = beacon.sf_pancoord
        self.assoc_permit = beacon.sf_assocpermit
        self.gts_permit = beacon.gts_spec_permit
        self.network_type = MACNetworkType.BEACON_ENABLED if self.beacon_order < 15 else MACNetworkType.NON_BEACON_ENABLED

    def __eq__(self, other):
        return self.coord_addr == other.coord_addr and self.coord_pan_id == other.coord_pan_id

    def __repr__(self):
        return (
                "Dot15d4PANNetwork("+
                "pan_id=" + hex(self.coord_pan_id)+", "
                "pan_coord=" + hex(self.coord_addr)+ ", "
                "channel_page=" + str(self.channel_page)+", "
                "channel=" + str(self.channel)+", "
                "link_quality=" + str(self.link_quality)+", "
                "network_type=" + (
                                    "beacon_enabled ({})".format(self.beacon_order) if
                                    self.network_type == MACNetworkType.BEACON_ENABLED else
                                    "non_beacon_enabled"
                )+", "
                "association=" + ("permitted" if self.assoc_permit else "forbidden") +
            ")"
        )



from whad.zigbee.stack.aps.constants import APSKeyAttribute, APSSecurityStatus
"""
Security-related classes for manipulating APS level cryptographic material.
"""

class APSTransportKeyData:
    """
    This class allows to transport security material from the APS to the ZDO layer.
    """
    def __init__(self, key):
        self.key = key


class APSTrustCenterLinkKeyData(APSTransportKeyData):
    """
    Specializes APSTransportKeyData class to store a Trust Center Link Key and its related data.
    """
    pass


class APSNetworkKeyData(APSTransportKeyData):
    """
    Specializes APSTransportKeyData class to store a Network Key and its related data.
    """
    def __init__(self, key, key_sequence_number, use_parent, parent_address=None):
        super().__init__(key)
        self.key_sequence_number = key_sequence_number
        self.use_parent = use_parent
        self.parent_address = parent_address


class APSApplicationLinkKeyData(APSTransportKeyData):
    """
    Specializes APSTransportKeyData class to store an Applicatino Key and its related data.
    """
    def __init__(self, key, partner_address):
        super().__init__(key)
        self.partner_address = partner_address


class APSKeyPair:
    """
    Implements a Key-Pair descriptor, needed to implement APS security features.
    """
    def __init__(
                    self,
                    device_address,
                    key,
                    key_attributes=APSKeyAttribute.PROVISIONAL_KEY,
                    outgoing_frame_counter=0,
                    incoming_frame_counter=0,
                    global_link_key=False
    ):
        self.device_address = device_address
        self.key = key
        self.key_attributes = key_attributes
        self.outgoing_frame_counter = outgoing_frame_counter
        self.incoming_frame_counter = incoming_frame_counter
        self.global_link_key = global_link_key

    def __repr__(self):
        if self.device_address is not None:
            return "APSKeyPair(device={:016x}, key={})".format(
                self.device_address, self.key
            )
        else:
            return "APSKeyPair(pre-installed, key={})".format(self.key)

class APSKeyPairSet:
    """
    Implement a set of APSKeyPair, providing an easy-to-use API to manipulate it.
    """
    def __init__(self, preinstalled_keys=[b"ZigBeeAlliance09"]):
        self.key_pair_set = []
        for preinstalled_key in preinstalled_keys:
            self.key_pair_set.append(
                APSKeyPair(
                    None,
                    preinstalled_key,
                    key_attributes=APSKeyAttribute.PROVISIONAL_KEY,
                    outgoing_frame_counter=0,
                    incoming_frame_counter=0,
                    global_link_key=True
                )
            )

    def select(self, address, unverified=True):
        matching_key_pairs = []
        preinstalled_keys = []
        for key_pair in self.key_pair_set:
            if (
                unverified or not unverified and
                key_pair.key_attributes in (
                                                APSKeyAttribute.PROVISIONAL_KEY,
                                                APSKeyAttribute.VERIFIED_KEY
                                            )
            ):
                if key_pair.device_address == address:
                    matching_key_pairs.append(key_pair)
                elif key_pair.device_address is None:
                    preinstalled_keys.append(key_pair)

        if len(matching_key_pairs) == 0:
            matching_key_pairs += preinstalled_keys

        return matching_key_pairs



"""WHAD mock device

This module provides various classes and helpers to implement mock devices
for WHAD:

- WhadDevice
"""

from .base import MockDevice
from .replay import ReplayMock
from .connector import MockConnector

__all__ = [
    "MockConnector",
    "MockDevice",
    "ReplayMock",
]



"""
Bluetooth Low Energy Tiny Stack for Unit Testing
================================================

GATT server profile:

- Generic Access Service (0x1800)
  - DeviceName (0x2A00), read/write permissions
- Battery Service (0x180F)
  - Battery Level (0x2A19), read/notify/indicate permissions
- Custom Service (6d02b600-1b51-4ef9-b753-1399e05debfd)
  - TX characteristic (6d02b601-1b51-4ef9-b753-1399e05debfd): writecmd
  - RX characteristic (6d02b602-1b51-4ef9-b753-1399e05debfd): read/notify permissions
"""
import logging
from struct import pack

from scapy.packet import Packet
from scapy.layers.bluetooth import ATT_Hdr

from whad.ble.profile.attribute import UUID

from .attribute import Attribute, PrimaryService, Characteristic, CharacteristicValue, \
    ClientCharacteristicConfigurationDescriptor

# ATT Procedures
from .procedure import UnexpectedProcError
from .read import ServerReadProcedure
from .readblob import ServerReadBlobProcedure
from .read_by_group_type import ServerReadByGroupTypeProcedure
from .read_by_type import ServerReadByTypeProcedure
from .find_info import ServerFindInformationProcedure, ServerFindByTypeValueProcedure
from .write import ServerWriteProcedure
from .writecmd import ServerWriteCommandProcedure

logger = logging.getLogger(__name__)

class GattServer:
    """Tiny GATT server"""

    def __init__(self, l2cap: 'Llcap' = None):
        """Initialize a GATT server."""
        ###
        # Initialize GATT attributes
        ###

        # Save L2CAP reference
        self.__l2cap = l2cap

        self.__attributes = [
            # Primary Service, Generic Access Service (0x1800), handle 1-3
            PrimaryService(1, 3,UUID(0x1800)),
            Characteristic(2, UUID(0x2A00), value_handle=3, properties=(
                Characteristic.PROP_READ | Characteristic.PROP_WRITE
            )),
            CharacteristicValue(3, UUID(0x2A00), b"EmulatedDevice"),

            # Primary Service, Battery Level Service (0x180F), handle 4-7
            PrimaryService(4, 7, UUID(0x180F)),
            Characteristic(5, UUID(0x2A19), value_handle=6, properties=(
                Characteristic.PROP_READ | Characteristic.PROP_NOTIFY
            )),
            CharacteristicValue(6, UUID(0x2A19), pack("<H", 0)),
            ClientCharacteristicConfigurationDescriptor(7),

            # Primary Service, custom service (6d02b602-1b51-4ef9-b753-1399e05debfd), handle 7-10
            PrimaryService(8, 13, UUID("6d02b600-1b51-4ef9-b753-1399e05debfd")),

            # TX Characteristic
            Characteristic(9, UUID("6d02b601-1b51-4ef9-b753-1399e05debfd"), value_handle=10, properties=(
                Characteristic.PROP_WRITE_WITHOUT_RESP
            )),
            CharacteristicValue(10, UUID("6d02b601-1b51-4ef9-b753-1399e05debfd"), b"\x00\x00\x00\x00",
                                write_without_resp=True),

            # RX Characteristic
            Characteristic(11, UUID("6d02b602-1b51-4ef9-b753-1399e05debfd"), value_handle=12, properties=(
                Characteristic.PROP_READ |  Characteristic.PROP_NOTIFY | Characteristic.PROP_INDICATE
            )),
            CharacteristicValue(12, UUID("6d02b602-1b51-4ef9-b753-1399e05debfd"), b"\x00\x00\x00\x00"),
            ClientCharacteristicConfigurationDescriptor(13),
        ]

        ###
        # Initialize GATT state
        ###

        # ATT MTU
        self.__mtu = 23

        # No current procedure
        self.__cur_procedure = None


        # Register procedures
        self.__procedures = [
            ServerReadProcedure,
            ServerReadBlobProcedure,
            ServerWriteProcedure,
            ServerWriteCommandProcedure,
            ServerReadByGroupTypeProcedure,
            ServerReadByTypeProcedure,
            ServerFindInformationProcedure,
            ServerFindByTypeValueProcedure,
        ]

    @property
    def attributes(self) -> list[Attribute]:
        """Server Attributes."""
        return self.__attributes

    def set_l2cap(self, obj):
        """Set L2CAP layer for this GattClient."""
        self.__l2cap = obj

    def on_pdu(self, request: Packet) -> list[Packet]:
        """Process incoming request."""
        # Find a suitable procedure if none in progress
        if self.__cur_procedure is None:
            # Find a procedure triggered by our request
            for proc in self.__procedures:
                if proc.trigger(request):
                    self.__cur_procedure = proc(self.attributes, self.__mtu)

        # If a suitable procedure has been found or is in progress, forward request.
        if self.__cur_procedure is not None:
            # Forward to current procedure
            answers: list[Packet] = self.__cur_procedure.process_request(request)

            # Unset current procedure if finished
            if self.__cur_procedure.done():
                self.__cur_procedure = None

            # Automatically add ATT_Hdr()
            if isinstance(answers, list):
                return [ ATT_Hdr()/ans for ans in answers ]
            else:
                # Should not happen so raise an exception
                raise UnexpectedProcError()
        else:
            logger.warning("[ble::mock::stack::gatt_server] No procedure to handle packet ! (%s)", request)
            return []




from whad.common.monitors import WhadMonitor
from scapy.all import conf
from scapy.layers.bluetooth4LE import *
from scapy.utils import PcapWriter,PcapReader
from scapy.error import Scapy_Exception
from os.path import exists
from whad.common.pcap import patch_pcap_metadata, extract_pcap_metadata
from time import time
from os import stat, remove
from stat import S_ISFIFO
from threading import Lock

import logging
logger = logging.getLogger(__name__)

class PcapWriterMonitor(WhadMonitor):
    """
    PcapWriterMonitor.

    Monitor allowing to export the traffic transmitted and received by the
    targeted connector to a PCAP file with appropriate header.

    :Usage:

        >>> monitor = PcapWriterMonitor("mypcapfile.pcap")
        >>> monitor.attach(connector)
        >>> monitor.start()

    """
    def __init__(self, pcap_file, monitor_reception=True, monitor_transmission=True):
        super().__init__(monitor_reception, monitor_transmission)
        self._pcap_file = pcap_file
        self._writer = None
        self._formatter = None
        self._reference_time = None
        self._start_time = None
        self._nb_pkts_written = 0
        self._writer_lock = Lock()


    @property
    def packets_written(self):
        '''Return the number of packets already written
        '''
        return self._nb_pkts_written


    def setup(self):
        # First, we check if the pcap already exists
        existing_pcap_file = exists(self._pcap_file)
        sync = False
        # If it exists, we have two cases:
        # - it is a named pipe and we have to provide sync=True to PcapWriter
        # - it is an already existing pcap file and we have to provide append=True to PcapWriter

        if existing_pcap_file:
            # Checks if it is a named FIFO
            if S_ISFIFO(stat(self._pcap_file).st_mode):
                logger.info("[i] Named pipe %s detected, syncing.", self._pcap_file)
                sync = True
            else:
                # Checks if it is an already existing pcap file.
                logger.info("[i] PCAP file %s exists, appending new packets.", self._pcap_file)
                try:
                    # We collect the first packet timestamp to use it as reference time
                    self._start_time = PcapReader(self._pcap_file).read_packet().time * 1000000
                except (EOFError, Scapy_Exception):
                    # Pcap is empty, remove it and open a new one
                    remove(self._pcap_file)
                    existing_pcap_file = False

        # Instantiate the PCAP Writer with the appropriate parameters
        self._writer_lock.acquire()
        self._writer = PcapWriter(
                                    self._pcap_file,
                                    append=existing_pcap_file and not sync,
                                    sync=sync
        )
        self._writer_lock.release()

        # Checks if there is a scapy packet formatter associated with the connector.
        # A formatter allows to describe manually how to build the packet, it is mainly
        # useful to populate a relevant header for PCAP export.
        self._formatter = self.default_formatter
        if (
            hasattr(self._connector, "format") and
            callable(getattr(self._connector, "format"))
        ):
            self._formatter = getattr(self._connector, "format")


    def close(self):
        # Acquire lock on writer
        self._writer_lock.acquire()
        if hasattr(self, "_writer") and self._writer is not None:


            # Close writer
            try:
                self._writer.close()
                patch_pcap_metadata(self._pcap_file, self._connector.domain)

            except BrokenPipeError:
                pass

            # Mark writer as not available anymore
            self._writer = None

        # Release lock on writer
        self._writer_lock.release()


    def default_formatter(self, packet):
        """
        Formatter used by default, if no formatter is found in the targeted connector.
        It only extracts the accurate timestamp if one is available in metadata.
        """
        if (
            hasattr(packet, "metadata") and
            hasattr(packet.metadata, "timestamp")
        ):
            return packet, packet.metadata.timestamp
        else:
            return packet, None

    def process_packet(self, packet):
        # Acquire lock on writer lock acquire multithread error

        self._writer_lock.acquire()

        if self._processing:
            # Note the current local clock timestamp in us
            now = time() * 1000000
            packet, timestamp = self._formatter(packet)
            
            # Relative time synchronization
            if timestamp is None:
                timestamp = now

            # Process accurate timestamp if available, else use local clock
            if self._reference_time is None:
                if self._start_time is None:
                    self._reference_time = (now, timestamp)
                else:
                    self._reference_time = (self._start_time, timestamp - (now - self._start_time))
                timestamp = now
            else:
                timestamp = self._reference_time[0] + (timestamp - self._reference_time[1])

            # Convert timestamp to second (float)
            packet.time = timestamp / 1000000
            try:
                if self._writer is not None:
                    # Write packet
                    self._writer.write(packet)
                    self._nb_pkts_written += 1
                else:
                    # We are trying to write to a closed PCAP monitor,
                    # issue a warning message
                    logger.warning('cannot write to PCAP: file has already been closed')
            except BrokenPipeError:
                pass
        # Release lock
        self._writer_lock.release()



from whad.hub.discovery import Domain, Capability
from whad.device import WhadDeviceConnector
from whad.helpers import message_filter
from whad.exceptions import UnsupportedDomain, UnsupportedCapability
from whad.hub.esb import ESBMetadata
#from whad.esb.connector.translator import ESBMessageTranslator
from whad.esb.esbaddr import ESBAddress
from whad.scapy.layers.esb import ESB_Hdr, ESB_Payload_Hdr, ESB_Ack_Response, ESB_Pseudo_Packet
from whad.scapy.layers.unifying import bind

from whad.hub.generic.cmdresult import Success, CommandResult
from whad.hub.unifying import EsbNodeAddress, Commands, RawPduReceived, PduReceived, UnifyingMetadata

class Unifying(WhadDeviceConnector):
    """
    Logitech Unifying protocol connector.

    This connector drives a Logitech Unifying (ESB-based) capable device with Unifying-specific WHAD messages.
    It is required by various role classes to interact with a real device and pre-process
    domain-specific messages.
    """
    domain = "unifying"

    def format(self, packet):
        """
        Converts a scapy packet with its metadata to a tuple containing a scapy packet with
        the appropriate header and the timestamp in microseconds.
        """
        return self.hub.unifying.format(packet)

    def __init__(self, device=None, synchronous=False):
        """
        Initialize the connector, open the device (if not already opened), discover
        the services (if not already discovered).

        If `auto` is set to False, PDUs must be processed manually and
        won't be forwarded to PDU-related callbacks.
        """
        self.__ready = False
        super().__init__(device)

        # Metadata cache
        self.__cached_channel = None
        self.__cached_address = None

        # Capability cache
        self.__can_send = None
        self.__can_send_raw = None

        # Open device and make sure it is compatible
        self.device.open()
        self.device.discover()

        # Check if device supports Logitech Unifying
        if not self.device.has_domain(Domain.LogitechUnifying):
            raise UnsupportedDomain("Logitech Unifying")
        else:
            self.__ready = True
            bind()


        # Set synchronous mode
        self.enable_synchronous(synchronous)

    def close(self):
        self.stop()
        self.device.close()

    def can_sniff(self):
        """
        Determine if the device implements a sniffer mode.
        """
        commands = self.device.get_domain_commands(Domain.LogitechUnifying)
        return (
            (commands & (1 << Commands.Sniff)) > 0 and
            (commands & (1 << Commands.Start))>0 and
            (commands & (1 << Commands.Stop))>0
        )

    def can_send(self):
        """
        Determine if the device can transmit packets.
        """
        if self.__can_send is None:
            commands = self.device.get_domain_commands(Domain.LogitechUnifying)
            self.__can_send = ((commands & (1 << Commands.Send))>0 or (commands & (1 << Commands.SendRaw)))
        return self.__can_send

    def send(self,pdu, address=None, channel=None, retransmission_count=1):
        """
        Send Logitech Unifying packets (on a single channel).
        """

        # Packet is not raw by default
        metadata = UnifyingMetadata()
        metadata.raw = False

        if not self.can_send():
            raise UnsupportedCapability("Send")
        # If we don't have address or channels, use the cached ones
        tx_address = address if address is not None else self.__cached_address
        tx_channel = channel if channel is not None else self.__cached_channel
        if self.support_raw_pdu():
            # If we support raw PDU but only got a payload, build a packet
            if ESB_Hdr not in pdu:
                packet = ESB_Hdr(address=tx_address) / pdu
            else:
                packet = pdu

            # Mark packet as raw
            metadata.raw = True

        # if we don't support raw PDU and got a packet, crop to keep only the payload
        elif ESB_Hdr in pdu:
            packet = pdu[ESB_Payload_Hdr:]
        # if we don't support raw PDU and got a payload, keep it as it is
        else:
            packet = pdu

        # Generate TX metadata
        packet.metadata = metadata
        packet.metadata.channel = tx_channel
        packet.metadata.address = tx_address

        if self.support_raw_pdu():
            # Set packet preamble depending on address
            if bytes.fromhex(packet.address[:2])[0] >= 0x80:
                packet.preamble = 0xAA
            else:
                packet.preamble = 0x55

        # Send packet
        ret = super().send_packet(packet)
        return ret

    def support_raw_pdu(self):
        """
        Determine if the device supports raw PDU.
        """
        if self.__can_send_raw is None:
            capabilities = self.device.get_domain_capability(Domain.LogitechUnifying)
            self.__can_send_raw = not (capabilities & Capability.NoRawData)
        return self.__can_send_raw


    def can_set_node_address(self):
        """
        Determine if the device can configure a Node address.
        """
        commands = self.device.get_domain_commands(Domain.LogitechUnifying)
        return (
            (commands & (1 << Commands.SetNodeAddress)) > 0
        )

    def start_sniff(self, channel=None, address="FF:FF:FF:FF:FF", show_acknowledgements=False):
        """
        Configure interface and start sniffing Logitech Unifying packets.
        """
        if not self.can_sniff():
            raise UnsupportedCapability("Sniff")

        self.__cached_address = ESBAddress(address)
        if channel is None:
            # Enable scanning mode
            channel = 0xFF
        else:
            self.__cached_channel = channel


        # Create a SniffMode message.
        msg = self.hub.unifying.create_sniff_mode(
            EsbNodeAddress(self.__cached_address.value),
            channel,
            show_acknowledgements
        )

        resp = self.send_command(msg, message_filter(CommandResult))
        return isinstance(resp, Success)


    def can_be_dongle(self):
        """
        Determine if the device implements a Logitech Dongle role mode.
        """
        commands = self.device.get_domain_commands(Domain.LogitechUnifying)
        return (
            (commands & (1 << Commands.LogitechDongleMode)) > 0 and
            (commands & (1 << Commands.Start))>0 and
            (commands & (1 << Commands.Stop))>0
        )

    def enable_dongle_mode(self, channel):
        """
        Enable Logitech Unifying dongle mode.
        """
        if not self.can_be_dongle():
            raise UnsupportedCapability("LogitechDongleMode")

        if channel is None:
            return False

        # Keep track of channel
        self.__cached_channel = channel

        # Create a DongleMode message
        msg = self.hub.unifying.create_dongle_mode(channel)

        resp = self.send_command(msg, message_filter(CommandResult))
        return isinstance(resp, Success)


    def can_be_keyboard(self):
        """
        Determine if the device implements a Logitech Keyboard role mode.
        """
        commands = self.device.get_domain_commands(Domain.LogitechUnifying)
        return (
            (commands & (1 << Commands.LogitechKeyboardMode)) > 0 and
            (commands & (1 << Commands.Start))>0 and
            (commands & (1 << Commands.Stop))>0
        )

    def enable_keyboard_mode(self, channel):
        """
        Enable Logitech Unifying Keyboard mode.
        """
        if not self.can_be_keyboard():
            raise UnsupportedCapability("LogitechKeyboardMode")

        if channel is None:
            # Enable scanning mode
            channel = 0xFF
        else:
            # Keep channel in cache
            self.__cached_channel = channel

        # Create a KeyboardMode message
        msg = self.hub.unifying.create_keyboard_mode(channel)

        resp = self.send_command(msg, message_filter(CommandResult))
        return isinstance(resp, Success)

    def can_be_mouse(self):
        """
        Determine if the device implements a Logitech Mouse role mode.
        """
        commands = self.device.get_domain_commands(Domain.LogitechUnifying)
        return (
            (commands & (1 << Commands.LogitechMouseMode)) > 0 and
            (commands & (1 << Commands.Start))>0 and
            (commands & (1 << Commands.Stop))>0
        )

    def enable_mouse_mode(self, channel):
        """
        Enable Logitech Unifying mouse mode.
        """
        if not self.can_be_mouse():
            raise UnsupportedCapability("LogitechMouseMode")

        if channel is None:
            # Enable scanning mode
            channel = 0xFF
        else:
            # Keep channel in cache
            self.__cached_channel = channel

        # Create a MouseMode message
        msg = self.hub.unifying.create_mouse_mode(channel)

        resp = self.send_command(msg, message_filter(CommandResult))
        return isinstance(resp, Success)


    def can_set_node_address(self):
        """
        Determine if the device can configure a Node address.
        """
        commands = self.device.get_domain_commands(Domain.LogitechUnifying)
        return (
            (commands & (1 << Commands.SetNodeAddress)) > 0
        )

    def set_node_address(self, address):
        """
        Select a specific node address.
        """
        if not self.can_set_node_address():
            raise UnsupportedCapability("SetNodeAddress")

        node_address = ESBAddress(address)
        # Keep address in cache
        self.__cached_address = node_address

        # Create a SetNodeAddress message.
        msg = self.hub.unifying.create_set_node_address(
            EsbNodeAddress(node_address.value)
        )

        resp = self.send_command(msg, message_filter(CommandResult))
        return isinstance(resp, Success)


    def can_sniff_pairing(self):
        """
        Determine if the device can follow a pairing procedure.
        """
        commands = self.device.get_domain_commands(Domain.LogitechUnifying)
        return (
            (commands & (1 << Commands.SniffPairing)) > 0
        )


    def sniff_pairing(self):
        """
        Follow a pairing procedure.
        """
        if not self.can_sniff_pairing():
            raise UnsupportedCapability("SniffPairing")

        # Create a SniffPairing message.
        msg = self.hub.unifying.create_sniff_pairing()

        resp = self.send_command(msg, message_filter(CommandResult))
        return isinstance(resp, Success)


    def start(self):
        """
        Start currently enabled mode.
        """
        # Create a Start message.
        msg = self.hub.unifying.create_start()

        resp = self.send_command(msg, message_filter(CommandResult))
        return isinstance(resp, Success)

    def stop(self):
        """
        Stop currently enabled mode.
        """
        # Create a Stop message.
        msg = self.hub.unifying.create_stop()

        resp = self.send_command(msg, message_filter(CommandResult))
        return isinstance(resp, Success)

    def on_discovery_msg(self, message):
        pass

    def on_generic_msg(self, message):
        pass

    def on_domain_msg(self, domain, message):
        pass

    def on_packet(self, packet):
        """Incoming packet handler.
        """
        if not self.__ready:
            return

        if packet.metadata.raw:
            self.on_raw_pdu(packet)
        else:
            self.on_pdu(packet)

    def on_event(self, event):
        """Incoming event handler.
        """
        pass

    def on_raw_pdu(self, packet):
        """Process incoming packet.
        """
        # Extract the PDU from raw packet
        if ESB_Payload_Hdr in packet:
            pdu = packet[ESB_Payload_Hdr:]
        else:
            pdu = ESB_Payload_Hdr()/ESB_Ack_Response()

        # Propagate metadata to PDU
        pdu.metadata = packet.metadata
        self.on_pdu(pdu)

    def on_pdu(self, packet):
        pass




"""ZigBee Packet translator
"""
import logging
from struct import pack

from scapy.layers.dot15d4 import Dot15d4, Dot15d4FCS

from whad.scapy.layers.dot15d4tap import Dot15d4TAP_Hdr, Dot15d4TAP_TLV_Hdr, Dot15d4TAP_FCS_Type
from whad.protocol.whad_pb2 import Message
from whad.hub import ProtocolHub
from whad.hub.dot15d4 import RawPduReceived, PduReceived, generate_dot15d4_metadata


logger = logging.getLogger(__name__)

class ZigbeeMessageTranslator(object):
    """Zigbee Whad message translator.

    This translator is used to provide the format of a specific scapy packet
    as well as standard methods to convert WHAD Zigbee messages into scapy packets
    (if it makes sense) and scapy packets into WHAD Zigbee messages.
    """

    def __init__(self, protocol_hub: ProtocolHub):
        self.__hub = protocol_hub

    def format(self, packet):
        """
        Converts a scapy packet with its metadata to a tuple containing a scapy packet with
        the appropriate header and the timestamp in microseconds.
        """
        if hasattr(packet, "metadata"):
            header, timestamp = packet.metadata.convert_to_header()
        else:
            header = Dot15d4TAP_Hdr()
            timestamp = None

        header.data.append(Dot15d4TAP_TLV_Hdr()/Dot15d4TAP_FCS_Type(
            fcs_type=int(Dot15d4FCS in packet)
            )
        )
        formatted_packet = header/packet
        return formatted_packet, timestamp

    def from_message(self, message):
        try:
            if isinstance(message, RawPduReceived):
                packet = Dot15d4FCS(bytes(message.pdu) + bytes(pack(">H", message.fcs)))
                packet.metadata = generate_dot15d4_metadata(message)
                return packet

            elif isinstance(message, PduReceived):
                packet = Dot15d4(bytes(message.pdu))
                packet.metadata = generate_dot15d4_metadata(message)
                return packet
        except AttributeError:
            return None


    def from_packet(self, packet, channel=11):
        msg = Message()

        if Dot15d4FCS in packet:
            # Create a SendPdu
            msg = self.__hub.dot15d4.create_send_raw_pdu(
                channel,
                bytes(packet)[:-2],
                packet.fcs
            )
        elif Dot15d4 in packet:
            msg = self.__hub.dot15d4.create_send_pdu(
                channel,
                bytes(packet)
            )
        else:
            # Raw MAC
            msg = self.__hub.dot15d4.create_send_pdu(
                channel,
                bytes(packet)
            )
        return msg



try:
    from setuptools import setup
    from setuptools.command.install import install
    from sys import platform, exit, executable
    from subprocess import Popen, DEVNULL, PIPE
    from os.path import exists, realpath
    from shutil import copy
    from os import listdir, geteuid, getlogin
    from grp import getgrall
except ImportError:
    print("Your operating system is not supported.")

class DevicesInstall(install):
    UDEV_LOCATION = [
        "/etc/udev/rules.d",
        "/lib/udev/rules.d",
        "/run/udev/rules.d",
        "/var/run/udev/rules.d"
    ]

    def install_udev_rule(self, name):
        print("Installing rules for %s device ..." % name)
        matching_rules = [udev_name for udev_name in self.udev_rules if name in udev_name]
        if len(matching_rules) == 0:
            if exists("resources/rules/"+name+".rules"):
                copy("resources/rules/"+name+".rules", "/usr/lib/udev/rules.d/40-"+name+".rules")
                return True
            else:
                print("Rule not found, skipping.")
                return False
        else:
            print("Rules detected for device %s (%s), skipping." % (name, ",".join(matching_rules)))
            return False

    def reload_udev_rules(self):
        print("Reloading udev rules...")
        process = Popen(['udevadm', 'control', '--reload-rules'], stderr=PIPE, stdout=PIPE)
        stdout, stderr = process.communicate()
        if len(stderr) > 0:
            return False

        process = Popen(['udevadm', 'trigger'],  stderr=PIPE, stdout=PIPE)
        stdout, stderr = process.communicate()
        if len(stderr) > 0:
            return False

        return True

    def install_serial_port_capabilities(self):
        print("Installing serial port capabilities...")
        # Find the right group which depends on the distribution (e.g. dialout
        # for Debian, uucp for Arch-Linux).
        groups = [ group.gr_name for group in getgrall() ]
        if "dialout" in groups:
            group = "dialout"
        elif "uucp" in groups:
            group = "uucp"
        else:
            raise Exception("Group granting serial port capabilities not found!")
        process = Popen(['usermod', '-a', '-G', group, getlogin()], stdout=PIPE, stderr=PIPE)
        stdout, stderr = process.communicate()

        if len(stderr) > 0:
            return False
        return True

    def install_hci_capabilities(self, interpreter):
        print("Installing HCI capabilities...")
        process = Popen(['getcap', interpreter], stdout=PIPE, stderr=PIPE)
        stdout, stderr = process.communicate()

        if "cap_net_admin,cap_net_raw=eip" not in stdout.decode('utf-8'):
            process = Popen(['setcap', 'cap_net_admin,cap_net_raw+eip',interpreter], stdout=PIPE, stderr=PIPE)
            stdout, stderr = process.communicate()

            if len(stdout) != 0 or len(stderr) != 0:
                return False
            return True
        else:
            return True

    def run(self):
        if platform == "linux" or platform == "linux2":
            if geteuid() != 0:
                print("You should run this installer as root !")
                exit(1)

            print("Installing devices...")
            # linux
            self.udev_rules = []
            for location in DevicesInstall.UDEV_LOCATION:
                if exists(location):
                    self.udev_rules += listdir(location)

            self.install_udev_rule("ubertooth")
            self.install_udev_rule("rzusbstick")
            self.install_udev_rule("rfstorm")
            self.install_udev_rule("yardstickone")

            if not self.reload_udev_rules():
                print("An error occured during udev rules reload.")

            python_interpreter = realpath(executable)
            if not self.install_hci_capabilities(python_interpreter):
                print("An error occured during HCI capabilities installation.")

            if not self.install_serial_port_capabilities():
                print("An error occured during Serial port capabilities installation.")
            else:
                print("User added to group 'dialout', please logout to take the modification into account.")
        else:
            print("Automatic device installation is not supported on your operating system.")
            exit(1)

if __name__ == "__main__":
    setup()



"""BLE emulate command handler
"""
from json import loads

from prompt_toolkit import print_formatted_text, HTML

from hexdump import hexdump

from whad.cli.app import command
from whad.ble import Peripheral, GenericProfile, AdvDataFieldList

class MonitoringProfile(GenericProfile):
    """GATT monitoring profile.
    """

    def on_characteristic_read(self, service, characteristic, offset=0, length=0):
        """Characteristic read hook.

        This hook is called whenever a characteristic is about to be read by a GATT client.
        If this method returns a byte array, this byte array will be sent back to the
        GATT client. If this method returns None, then the read operation will return an
        error (not allowed to read characteristic value).
        

        :param BlePrimaryService service: Service owning the characteristic
        :param BleCharacteristic characteristic: Characteristic object
        :param int offset: Read offset (default: 0)
        :param int length: Max read length
        :return: Value to return to the GATT client
        """
        print_formatted_text(HTML((
            f"<ansigreen>Reading</ansigreen> characteristic "
            f"<ansicyan>{characteristic.uuid}</ansicyan>"
            f" of service <ansicyan>{service.uuid}</ansicyan>"
        )))
        hexdump(characteristic.value)

    def on_connect(self, conn_handle):
        print_formatted_text(HTML(f"<ansired>New connection</ansired> handle:{conn_handle}"))

    def on_disconnect(self, conn_handle):
        print_formatted_text(HTML(f"<ansired>Disconnection</ansired> handle:{conn_handle}"))

    def on_characteristic_written(self, service, characteristic, offset=0, value=b'',
                                  without_response=False):
        """Characteristic written hook

        This hook is called whenever a charactertistic has been written by a GATT
        client.
        """
        print_formatted_text(HTML((
            f"<ansimagenta>Wrote</ansimagenta> to characteristic"
            f"<ansicyan>{characteristic.uuid}</ansicyan>"
            f" of service <ansicyan>{service.uuid}</ansicyan>"
        )))
        hexdump(value)


    def on_characteristic_subscribed(self, service, characteristic, notification=False,
                                     indication=False):
        # Check if we have a hook to call
        print_formatted_text(HTML((
            f"<ansicyan>Subscribed</ansicyan> to characteristic"
            f"<ansicyan>{characteristic.uuid}</ansicyan>"
            f" of service <ansicyan>{service.uuid}</ansicyan>"
        )))

    def on_characteristic_unsubscribed(self, service, characteristic):
        print_formatted_text(HTML((
            f"<ansicyan>Unsubscribed</ansicyan> to characteristic"
            f"<ansicyan>{characteristic.uuid}</ansicyan>"
            f" of service <ansicyan>{service.uuid}</ansicyan>"
        )))

    def on_notification(self, service, characteristic, value):
        print_formatted_text(HTML((
            f"<ansicyan>Notification</ansicyan> from characteristic"
            f"<ansicyan>{characteristic.uuid}</ansicyan>"
            f" of service <ansicyan>{service.uuid}</ansicyan>"
        )))
        hexdump(value)

    def on_indication(self, service, characteristic, value):
        print_formatted_text(HTML((
            f"<ansicyan>Indication</ansicyan> from characteristic"
            f"<ansicyan>{characteristic.uuid}</ansicyan>"
            f" of service <ansicyan>{service.uuid}</ansicyan>"
        )))
        hexdump(value)

def check_profile(profile):
    """Check profile validity.

    :param dict p: Profile
    :return bool: True if profile format is OK, False otherwise
    """
    # Make sure we have a 'devinfo' entry
    if 'devinfo' not in profile:
        return False

    # Make sure we have a valid adv_data entry
    if 'adv_data' not in profile['devinfo']:
        return False
    try:
        bytes.fromhex(profile["devinfo"]["adv_data"])
    except ValueError:
        return False

    # Make sure we have a valid scan_rsp entry (if provided)
    if 'scan_rsp' not in profile['devinfo']:
        return False
    if profile['devinfo']['scan_rsp'] is not None:
        try:
            bytes.fromhex(profile["devinfo"]["scan_rsp"])
        except ValueError:
            return False

    # Make sure we have a BD address
    if 'bd_addr' not in profile['devinfo']:
        return False

    # OK
    return True



@command('emulate')
def emulate_handler(app, command_args):
    """emulate a device from an exported profile
    
    <ansicyan><b>emulate</b> <i>[profile JSON file]</i></ansicyan>

    This command will emulate a device from its exported GATT profile (using
    the <ansicyan>profile</ansicyan> command). It will try to spoof its BD address (if possible),
    and will expose the exact same services and characteristics as the original one.

    If a GATT client connects to the emulated device, all the operations will
    be logged and detailed.
    """
    # We need to have an interface specified
    if app.interface is not None:
        if len(command_args) >= 1:
            # Import profile JSON
            with open(command_args[0],'r', encoding="utf-8") as dev_profile:
                profile_json = dev_profile.read()
                profile = loads(profile_json)

            # Check profile and emulate if everything is OK
            if check_profile(profile):
                # Load device info
                device_adv_data = bytes.fromhex(profile["devinfo"]["adv_data"])
                device_scan_rsp = bytes.fromhex(profile["devinfo"]["scan_rsp"])

                # Create a profile
                device_profile = MonitoringProfile(from_json = profile_json)
                periph = Peripheral(
                    app.interface,
                    profile=device_profile,
                    adv_data=AdvDataFieldList.from_bytes(device_adv_data),
                    scan_data=AdvDataFieldList.from_bytes(device_scan_rsp),
                    bd_address=profile['devinfo']['bd_addr']
                )
                periph.start()
                print((f"Emulation of device {profile['devinfo']['bd_addr']} is active,"
                       f" press any key to stop..."))
                try:
                    input()
                except KeyboardInterrupt:
                    pass
                print("\rStopping emulation ...")
                periph.stop()
            else:
                app.error("Bad JSON file format")
        else:
            app.error("You must provide an exported profile file (JSON).")



from whad.zigbee.crypto import TouchlinkKeyManager, TransportKeyDistribution


analyzers = {
    "touchlink_key_cracking" : TouchlinkKeyManager,
    "transport_key_cracking" : TransportKeyDistribution
}



from Cryptodome.Cipher import AES
from struct import pack
from copy import copy

from whad.unifying.exceptions import MissingEncryptedKeystrokePayload
from whad.common.analyzer import TrafficAnalyzer
from whad.scapy.layers.esb import ESB_Hdr
from whad.scapy.layers.unifying import bind, Logitech_Unifying_Hdr, \
    Logitech_Unencrypted_Keystroke_Payload, Logitech_Encrypted_Keystroke_Payload, \
    Logitech_Pairing_Request_1_Payload, Logitech_Pairing_Request_2_Payload, \
    Logitech_Pairing_Response_1_Payload, Logitech_Pairing_Response_2_Payload


class LogitechUnifyingCryptoManager:
    def __init__(self, key):
        self.key = key

    def e(self, input_data):
        aes = AES.new(self.key, AES.MODE_ECB)
        return aes.encrypt(input_data)[:8]

    def xor(self, a, b):
        result = []
        for i in range(len(a)):
            result += [a[i] ^ b[i]]
        return bytes(result)

    def extractPayload(self, packet):
        return packet.hid_data + bytes([packet.unknown])

    def generateAESInputData(self, counter):
        return bytes([0x04, 0x14, 0x1d, 0x1f, 0x27, 0x28, 0x0d]) + pack(">I", counter) + bytes([0x0a, 0x0d, 0x13, 0x26, 0x0e])

    def decrypt(self, packet):
        if isinstance(packet, bytes):
            packet = Logitech_Unifying_Hdr(packet)

        elif isinstance(packet, ESB_Hdr):
            packet = packet[Logitech_Unifying_Hdr:]

        if Logitech_Encrypted_Keystroke_Payload not in packet:
            raise MissingEncryptedKeystrokePayload()

        ciphertext = self.extractPayload(packet)
        counter = packet.aes_counter

        aes_in = self.generateAESInputData(counter)
        vector = self.e(aes_in)

        result = self.xor(vector, ciphertext)

        decrypted_packet = copy(packet)
        decrypted_packet.hid_data = result[:7]
        decrypted_packet.unknown = result[7]
        decrypted_packet.aes_counter = counter
        return decrypted_packet

    def encrypt(self, packet):
        if isinstance(packet, bytes):
            packet = Logitech_Unifying_Hdr(packet)

        elif isinstance(packet, ESB_Hdr):
            packet = packet[Logitech_Unifying_Hdr:]

        if Logitech_Encrypted_Keystroke_Payload not in packet:
            raise MissingEncryptedKeystrokePayload()

        plaintext = self.extractPayload(packet)
        counter = packet.aes_counter

        aes_in = self.generateAESInputData(counter)
        vector = self.e(aes_in)

        result = self.xor(vector, plaintext)

        encrypted_packet = copy(packet)
        encrypted_packet.hid_data = result[:7]
        encrypted_packet.unknown = result[7]
        encrypted_packet.aes_counter = counter
        return encrypted_packet

class LogitechUnifyingKeyDerivation(TrafficAnalyzer):
    def __init__(self, address=None, dongle_wpid=None, device_wpid=None, dongle_nonce=None, device_nonce=None):
        self.reset()
        self.address = address
        self.dongle_wpid = dongle_wpid
        self.device_wpid = device_wpid
        self.dongle_nonce = dongle_nonce
        self.device_nonce = device_nonce

    def process_packet(self, packet):
        if Logitech_Pairing_Request_1_Payload in packet:
            self.trigger()
            self.device_wpid = pack(">H", packet.device_wpid)
            self.mark_packet(packet)
        elif Logitech_Pairing_Response_1_Payload in packet:
            self.address = bytes.fromhex(packet.rf_address.replace(":", ""))
            self.dongle_wpid = pack(">H", packet.dongle_wpid)
            self.mark_packet(packet)
        elif Logitech_Pairing_Request_2_Payload in packet:
            self.device_nonce = packet.device_nonce
            self.mark_packet(packet)
        elif Logitech_Pairing_Response_2_Payload in packet:
            self.dongle_nonce = packet.dongle_nonce
            self.mark_packet(packet)
        if (
                self.address is not None and
                self.dongle_wpid is not None and
                self.device_wpid is not None and
                self.dongle_nonce is not None and
                self.device_nonce is not None
        ):
            self.complete()

    @property
    def output(self):
        return {"key" : self.key}

    @property
    def key(self):
        if (
                self.address is not None and
                self.dongle_wpid is not None and
                self.device_wpid is not None and
                self.dongle_nonce is not None and
                self.device_nonce is not None
        ):
            return self._generate_key()
        else:
            return None

    def reset(self):
        super().reset()
        self.address = None
        self.dongle_wpid = None
        self.device_wpid = None
        self.dongle_nonce = None
        self.device_nonce = None

    def _generate_key(self):
        raw_key_material = self.address[:4] + self.device_wpid + self.dongle_wpid + self.device_nonce + self.dongle_nonce
        key = [0 for _ in range(16)]

        key[0] = raw_key_material[7]
        key[1] = raw_key_material[1] ^ 0xff
        key[2] = raw_key_material[0]
        key[3] = raw_key_material[3]
        key[4] = raw_key_material[10]
        key[5] = raw_key_material[2] ^ 0xff
        key[6] = raw_key_material[9] ^ 0x55
        key[7] = raw_key_material[14]
        key[8] = raw_key_material[8]
        key[9] = raw_key_material[6]
        key[10] = raw_key_material[12] ^ 0xff
        key[11] = raw_key_material[5]
        key[12] = raw_key_material[13]
        key[13] = raw_key_material[15] ^ 0x55
        key[14] = raw_key_material[4]
        key[15] = raw_key_material[11]

        return bytes(key)

class LogitechUnifyingDecryptor:
    def __init__(self, *keys):
        self.keys = list(keys)

    def add_key(self, key):
        if isinstance(key, str):
            if len(key) == 16:
                key = key.encode('ascii')
            else:
                try:
                    key = bytes.fromhex(key.replace(":",""))
                except ValueError:
                    return False

        if not isinstance(key, bytes) or len(key) != 16:
            return False

        if key not in self.keys:
            self.keys.append(key)
            return True
        return False

    def attempt_to_decrypt(self, packet):
        if Logitech_Encrypted_Keystroke_Payload not in packet:
            raise MissingEncryptedKeystrokePayload()

        for key in self.keys:
            manager = LogitechUnifyingCryptoManager(key)
            decrypted = manager.decrypt(packet)
            
            if b"\x00\x00" in decrypted.hid_data:
                return decrypted[Logitech_Unifying_Hdr:], True

        return packet[Logitech_Unifying_Hdr:], False



"""WHAD Protocol PHY packet messages abstraction layer.
"""
from whad.protocol.whad_pb2 import Message
from whad.scapy.layers.phy import Phy_Packet
from whad.hub.phy import PhyMetadata, Endianness, Modulation, Syncword
from ..message import pb_bind, PbFieldInt, PbFieldBytes,PbFieldArray, PbMessageWrapper, \
    dissect_failsafe
from . import PhyDomain

from .timestamp import Timestamp

@pb_bind(PhyDomain, 'datarate', 1)
class SetDatarate(PbMessageWrapper):
    """PHY Datarate setting message
    """

    rate = PbFieldInt('phy.datarate.rate')

@pb_bind(PhyDomain, 'endianness', 1)
class SetEndianness(PbMessageWrapper):
    """PHY packet endianness setting message
    """

    endianness = PbFieldInt('phy.endianness.endianness')

@pb_bind(PhyDomain, 'tx_power', 1)
class SetTxPower(PbMessageWrapper):
    """PHY TX power configuration message
    """

    power = PbFieldInt('phy.tx_power.tx_power')

@pb_bind(PhyDomain, 'packet_size', 1)
class SetPacketSize(PbMessageWrapper):
    """PHY packet size configuration message
    """

    packet_size = PbFieldInt('phy.packet_size.packet_size')

@pb_bind(PhyDomain, 'sync_word', 1)
class SetSyncWord(PbMessageWrapper):
    """PHY sync word configuration message
    """

    sync_word = PbFieldBytes('phy.sync_word.sync_word')


@pb_bind(PhyDomain, 'send', 1)
class SendPacket(PbMessageWrapper):
    """PHY packet send message
    """

    packet = PbFieldBytes('phy.send.packet')

    @dissect_failsafe
    def to_packet(self):
        """Convert message to packet
        """
        return Phy_Packet(self.packet)


    @staticmethod
    def from_packet(packet):
        """Convert packet to message
        """
        return SendPacket(
            packet=bytes(packet)
        )

@pb_bind(PhyDomain, 'send_raw', 1)
class SendRawPacket(PbMessageWrapper):
    """PHY raw packet send message
    """

    iq = PbFieldArray('phy.send_raw.iq')

    @dissect_failsafe
    def to_packet(self):
        """Convert message to packet
        """
        return Phy_Packet(self.packet)


    @staticmethod
    def from_packet(packet):
        """Convert packet to message
        """
        return SendRawPacket(
            iq=[]
        )

@pb_bind(PhyDomain, 'packet', 1)
class PacketReceived(PbMessageWrapper):
    """PHY packet received notification message
    """

    frequency = PbFieldInt('phy.packet.frequency')
    packet = PbFieldBytes('phy.packet.packet')
    rssi = PbFieldInt('phy.packet.rssi', optional=True)
    timestamp = PbFieldInt('phy.packet.timestamp', optional=True)

    @dissect_failsafe
    def to_packet(self):
        """Convert message to packet
        """
        packet = Phy_Packet(self.packet)
        packet.metadata = PhyMetadata()
        packet.metadata.frequency = self.frequency
        packet.metadata.raw = False
        if self.rssi is not None:
            packet.metadata.rssi = self.rssi
        if self.timestamp is not None:
            packet.metadata.timestamp = self.timestamp

        return packet

    @staticmethod
    def from_packet(packet):
        """Convert packet to message
        """
        msg = PacketReceived(
            frequency=packet.metadata.frequency,
            packet=bytes(packet)
        )
        if packet.metadata.rssi is not None:
            msg.rssi = packet.metadata.rssi
        if packet.metadata.timestamp is not None:
            msg.timestamp = packet.metadata.timestamp

        return msg


@pb_bind(PhyDomain, 'packet', 2)
class ExtendedPacketReceived(PacketReceived):
    """PHY packet received notification message (extended - v2)
    """

    frequency = PbFieldInt('phy.packet.frequency')
    packet = PbFieldBytes('phy.packet.packet')
    rssi = PbFieldInt('phy.packet.rssi', optional=True)
    timestamp = PbFieldInt('phy.packet.timestamp', optional=True)
    syncword = PbFieldBytes('phy.packet.syncword')
    deviation = PbFieldInt('phy.packet.deviation')
    datarate = PbFieldInt('phy.packet.datarate')
    endian = PbFieldInt('phy.packet.endian')
    modulation = PbFieldInt('phy.packet.modulation')

    @dissect_failsafe
    def to_packet(self):
        """Convert message to packet
        """
        packet = Phy_Packet(self.packet)
        packet.metadata = PhyMetadata()
        packet.metadata.frequency = self.frequency
        packet.metadata.raw = False
        if self.rssi is not None:
            packet.metadata.rssi = self.rssi
        if self.timestamp is not None:
            packet.metadata.timestamp = self.timestamp
        if self.syncword is not None:
            packet.metadata.syncword = Syncword(self.syncword)
        if self.deviation is not None:
            packet.metadata.deviation = self.deviation
        if self.datarate is not None:
            packet.metadata.datarate = self.datarate
        if self.endian is not None:
            packet.metadata.endianness = Endianness(self.endian)
        if self.modulation is not None:
            packet.metadata.modulation = Modulation(self.modulation)

        return packet

    @staticmethod
    def from_packet(packet):
        """Convert packet to message
        """
        msg = ExtendedPacketReceived(
            frequency=packet.metadata.frequency,
            packet=bytes(packet)
        )
        if packet.metadata.rssi is not None:
            msg.rssi = packet.metadata.rssi
        if packet.metadata.timestamp is not None:
            msg.timestamp = packet.metadata.timestamp

        if packet.metadata.endianness is not None:
            msg.endian = packet.metadata.endianness

        if packet.metadata.datarate is not None:
            msg.datarate = packet.metadata.datarate

        if packet.metadata.deviation is not None:
            msg.deviation = int(packet.metadata.deviation)

        if packet.metadata.modulation is not None:
            msg.modulation = int(packet.metadata.modulation)

        if packet.metadata.syncword is not None:
            msg.syncword = bytes(packet.metadata.syncword)

        return msg

@pb_bind(PhyDomain, 'raw_packet', 1)
class RawPacketReceived(PbMessageWrapper):
    """PHY packet received notification message

    IQ not supported yet.
    """

    frequency = PbFieldInt('phy.raw_packet.frequency')
    packet = PbFieldBytes('phy.raw_packet.packet')
    rssi = PbFieldInt('phy.raw_packet.rssi', optional=True)
    iq = PbFieldArray('phy.raw_packet.iq')
    timestamp = PbFieldInt('phy.raw_packet.timestamp', optional=True)

    @dissect_failsafe
    def to_packet(self):
        """Convert message to packet
        """
        packet = Phy_Packet(self.packet)
        packet.metadata = PhyMetadata()
        packet.metadata.frequency = self.frequency
        packet.metadata.raw = True
        if self.rssi is not None:
            packet.metadata.rssi = self.rssi
        if self.timestamp is not None:
            packet.metadata.timestamp = self.timestamp

        return packet

    @staticmethod
    def from_packet(packet):
        """Convert packet to message
        """
        msg = PacketReceived(
            frequency=packet.metadata.frequency,
            packet=bytes(packet)
        )
        if packet.metadata.rssi is not None:
            msg.rssi = packet.metadata.rssi
        if packet.metadata.timestamp is not None:
            msg.timestamp = packet.metadata.timestamp

        return msg



@pb_bind(PhyDomain, 'raw_packet', 2)
class ExtendedRawPacketReceived(PbMessageWrapper):
    """PHY packet received notification message (extended - v2)

    IQ not supported yet.
    """

    frequency = PbFieldInt('phy.raw_packet.frequency')
    packet = PbFieldBytes('phy.raw_packet.packet')
    rssi = PbFieldInt('phy.raw_packet.rssi', optional=True)
    iq = PbFieldArray('phy.raw_packet.iq')
    timestamp = PbFieldInt('phy.raw_packet.timestamp', optional=True)
    syncword = PbFieldBytes('phy.packet.syncword')
    deviation = PbFieldInt('phy.packet.deviation')
    datarate = PbFieldInt('phy.packet.datarate')
    endianness = PbFieldInt('phy.packet.endian')
    modulation = PbFieldInt('phy.packet.modulation')

    @dissect_failsafe
    def to_packet(self):
        """Convert message to packet
        """
        packet = Phy_Packet(self.packet)
        packet.metadata = PhyMetadata()
        packet.metadata.frequency = self.frequency
        packet.metadata.raw = True
        if self.rssi is not None:
            packet.metadata.rssi = self.rssi
        if self.timestamp is not None:
            packet.metadata.timestamp = self.timestamp
        if self.syncword is not None:
            packet.metadata.syncword = Syncword(self.syncword)
        if self.deviation is not None:
            packet.metadata.deviation = self.deviation
        if self.datarate is not None:
            packet.metadata.datarate = self.datarate
        if self.endian is not None:
            packet.metadata.endianness = Endianness(self.endian)
        if self.modulation is not None:
            packet.metadata.modulation = Modulation(self.modulation)

        return packet

    @staticmethod
    def from_packet(packet):
        """Convert packet to message
        """
        msg = PacketReceived(
            frequency=packet.metadata.frequency,
            packet=bytes(packet)
        )
        if packet.metadata.rssi is not None:
            msg.rssi = packet.metadata.rssi
        if packet.metadata.timestamp is not None:
            msg.timestamp = packet.metadata.timestamp

        if packet.metadata.endianness is not None:
            msg.endian = endianness

        if packet.metadata.datarate is not None:
            msg.datarate = datarate

        if packet.metadata.deviation is not None:
            msg.deviation = deviation

        if packet.metadata.modulation is not None:
            msg.modulation = modulation

        if packet.metadata.syncword is not None:
            msg.syncword = syncword
        return msg
